; ******************************************************************************
; *                                                                            *
; *                              "SUB DESTROYER"                               *
; *                                                                            *
; *                   A MiniGame for the Commodore 64 (PAL)                    *
; *                                                                            *
; *                     Source Code (C) Geir Straume 2004                      *
; *                                                                            *
; *       This source file may be assembled using AS65, a 6502 assembler       *
; *       by Frank A. Vorstenbosch. AS65 is available for download at          *
; *       http://www.kingswood-consulting.co.uk/assemblers/                    *
; *                                                                            *
; *       To assemble:  as65 -qvosubdestroyer.prg subdestroyer.asm             *
; *                                                                            *
; *                                                                            *
; *                                                                            *
; *       geir.straume@chello.no                                               *
; *       http://home.chello.no/~gstraume/c64/                                 *
; *                                                                            *
; ******************************************************************************

startaddr  equ  $0801              ; Start address of program

           org  startaddr-2        ;
           dw   startaddr          ; First 2 bytes of file is the start address

; ******************************************************************************
; *  BASIC stub (SYS 2061)                                                     *
; ******************************************************************************

           db   11,8,0,0,158,50,48,54,49,0,0,0

; ******************************************************************************
; *  Addresses of tables, chip registers, variables etc.                       *
; ******************************************************************************

irqvector  equ  $0314              ; Lo and hi byte of raster interrupt routine
nmivector  equ  $0318              ; Lo and hi byte of non-maskable interrupt
v          equ  $d000              ; VIC chip
s          equ  $d400              ; SID chip
joystick   equ  $dc00              ; Joystick port 2

tmpspdata  equ  $7800              ; Temporary area for sprite data
charset    equ  $5200              ; Bit patterns for alphanumeric characters
vicidxw    equ  $6000              ; VIC register indexes for write
vicvalw    equ  vicidxw+256        ; VIC register values to be written
vicidxr    equ  vicvalw+256        ; VIC register indexes for read
vicvalr    equ  vicidxr+256        ; VIC register values read
vicidxwtop equ  vicidxw+211        ; VIC register indexes for top sprite row
vicidxwbtm equ  vicidxwtop+19      ; VIC register indexes for bottom sprite row
vicvalwtop equ  vicvalw+211        ; VIC register values for top sprite row
vicvalwbtm equ  vicvalwtop+19      ; VIC register values for bottom sprite row
titlespr   equ  $6400              ; Sprite pointer data for title screen
gamespr    equ  titlespr+256       ; Sprite pointer data for game screen
mspdatalo  equ  gamespr+80         ; Lo byte of mine sprite image addresses
mspdatahi  equ  mspdatalo+20       ; Hi byte of mine sprite image addresses
vicvalwidx equ  mspdatahi+20       ; Indexes to 'vicvalw' table
vicvalwval equ  vicvalwidx+13      ; Values for 'vicvalw' table
sptxtdata  equ  $6800              ; Text data
hiscore    equ  sptxtdata+84       ; High score
score      equ  hiscore+6          ; Score
mines      equ  score+6            ; Blank char + number of mines (3 bytes)
pts        equ  mines+3            ; Points to add to score (5 bytes)
level      equ  pts+5              ; Blank char + level number (3 bytes)

new_addr   equ  71                 ; Special code used in data to be relocated
endofdata  equ  72                 ; Special code used in data to be relocated

; Zero page variables
; * indicates that a variable MUST be located at the specified zero page
;   address because a ROM routine is used to increment it

mspxpos    equ  2                  ; X-coordinates of mine sprites (20 bytes)
mspxposmsb equ  22                 ; X-coordinates MSB of mine sprites (20)
gstatus    equ  42                 ; Game status
leveldata  equ  43                 ; Address of current level data byte (2) *
mspminecnt equ  45                 ; Number of mines within each sprite (20)
minespnum  equ  65                 ; Mine sprite number used (10)
mineypos   equ  75                 ; Y-coordinate of mine within sprite (10)
minespoffs equ  85                 ; Byte offset within sprite image (10)
minebitidx equ  95                 ; Index to bit tables for drawing mine (10)
mineupdcnt equ  105                ; Counter for update of mines on screen
minetabidx equ  106                ; Index to mine tables
minenum    equ  107                ; Mine number
freezedsp  equ  108                ; Bit 6 set if display is freezed
objxpos    equ  109                ; X-coord. of 18 subs + ship (19)
objxposmsb equ  128                ; X-coord. MSB of 18 subs + ship (19)
objexpcnt  equ  147                ; Explosion count for 18 subs + ship (19)
framecnt   equ  166                ; Decreases once per screen frame
tmpa       equ  167                ; Temporary variable
tmpb       equ  168                ; Temporary variable
tmpc       equ  169                ; Temporary variable
tmpd       equ  170                ; Temporary variable
tmpcnt     equ  171                ; Temporary counter variable (non-irq)
va         equ  172                ; Temporary vector variable (2) *
vb         equ  174                ; Temporary vector variable (2) (non-irq)
subrelmode equ  176                ; Current submarine release mode
subrelint  equ  177                ; Sub release interval counter
subrelval  equ  178                ; Sub release interval value
subrelcnt  equ  179                ; Sub release counter
ships      equ  180                ; Number of ships left
dropdelay  equ  181                ; Delay counter for dropping mines
islastrow  equ  182                ; Bit 7 set if mine is in last row (non-irq)
islastr    equ  183                ; Bit 7 set if mine is in last row (irq)
txttabidx  equ  184                ; Index to tables for writing text in sprites
fastsub    equ  185                ; Index of sub which moves every frame
torpedosub equ  186                ; Index of sub which may launch a torpedo
torpedocnt equ  187                ; Torpedo movement counter
torpspimg  equ  188                ; Address of torpedo sprite image (2)
torpspoffs equ  190                ; Byte offset of torpedo within sprite image
disablecnt equ  191                ; Number of mines disabled by collision check
isshiplost equ  192                ; Non-zero if a ship has just been lost
levelnum   equ  193                ; Level number in BCD format
endoflevel equ  194                ; Bit 7 set at the end of a level
msgcounter equ  195                ; Counter for display of "LEVEL n" message

zpdata     equ  196                ; Start address of various zero page data
andmask2   equ  zpdata             ; Bit table for masking v+16 values
andmask1   equ  andmask2+2         ; Bit table for masking v+16 values
andmask3   equ  andmask1+4         ; Bit table for masking v+16 values
oramask2   equ  andmask3+4         ; Bit table for setting v+16 values
oramask1   equ  oramask2+2         ; Bit table for general use
mineandval equ  oramask1+8         ; Bit table for erasing mine
mineoraval equ  mineandval+8       ; Bit table for drawing mine
txtstart   equ  mineoraval+8       ; Start indexes to table 'sptxtdata'
txtend     equ  txtstart+6         ; End indexes to table 'sptxtdata'
txtsplo    equ  txtend+6           ; Lo byte of text sprite image address
txtsphi    equ  txtsplo+6          ; Hi byte of text sprite image address

; ******************************************************************************
; *                 S t a r t   o f   p r o g r a m   c o d e                  *
; ******************************************************************************

main       jsr  setup              ; Initial setup

titlescr   jsr  dsptitle           ; Display title screen

           lda  #16                ;
ma1        jsr  firebutton         ;
           beq  ma1                ; Wait for player to press fire button

           jsr  dofreeze           ; Freeze display and wait for acknowledgement
           jsr  newgame            ; Set up a new game
           lsr  gstatus            ; Set status to "Game in progress"
           lsr  freezedsp          ; Do not freeze display anymore

; Move mines during game and wait for status to change
ma2        bit  gstatus            ; Check game status
           bvc  gameover           ; Status has changed
           lda  mineupdcnt         ; Move mines now?
           beq  ma2                ; No
           jsr  movemines          ; Yes
           dec  mineupdcnt         ;
           bpl  ma2                ; Branch always

; ******************************************************************************
; *  Game over                                                                 *
; ******************************************************************************

gameover   ldx  framecnt           ;
           inx                     ;
ma3        cpx  framecnt           ; Wait a few seconds
           bne  ma3                ;
           jsr  dofreeze           ; Freeze display and wait for acknowledgement
           bpl  titlescr           ; Branch always

; ******************************************************************************
; *  Check fire button                                                         *
; ******************************************************************************

firebutton lda  joystick           ; Bit 0=up, 1=down, 2=left, 3=right, 4=fire
           and  #%00010000         ;
fb1        cmp  #0                 ; Compare to previous fire button status
           beq  fb2                ; Prevent autofire
           sta  fb1+1              ; Self-modifying code
           eor  #%00010000         ;
fb2        rts

; ******************************************************************************
; *  Freeze display                                                            *
; ******************************************************************************

dofreeze   lda  #%11000000         ; Bit 6=freeze, bit 7=get ack from irq
           sta  freezedsp          ; Tell irq routines to freeze display
df1        bit  freezedsp          ;
           bmi  df1                ; Wait for acknowledgement from irq
           rts

; ******************************************************************************
; *              R a s t e r   i n t e r r u p t   r o u t i n e s             *
; ******************************************************************************

; ******************************************************************************
; *  The first interrupt                                                       *
; ******************************************************************************

irqa       lda  #lo irqb           ;
           sta  irqvector          ; Set lo byte of next interrupt routine
           lda  #11                ;
           sta  v+17               ; Display sprites, but no text screen
           nop                     ;
           nop                     ;
           nop                     ;
           nop                     ;
           nop                     ;
           nop                     ;
           inc  v+18               ; Set raster line for next interrupt
           dec  v+25               ; Acknowledge current interrupt
           cli                     ;
           ldy  #8                 ; Timing
ia1        dey                     ;
           bpl  ia1                ;
           nop                     ; Next raster interrupt will occur while
           nop                     ; executing one of these instructions
           nop                     ;
           nop                     ;
           jmp  $ea81              ;

; ******************************************************************************
; *  The second interrupt                                                      *
; ******************************************************************************

; The second interrupt will be off by max 1 cycle
irqb       lda  #lo irqc           ;
           sta  irqvector          ; Set lo byte of next interrupt routine
           lda  #hi irqc           ;
           sta  irqvector+1        ; Set hi byte of next interrupt routine
           nop                     ;
           bit  0                  ; Timing
           lda  v+18               ;
           cmp  v+18               ;
           beq  ib1                ; Final raster synchronization

ib1        ldy  #5                 ;
ib2        dey                     ; Timing
           bpl  ib2                ;
           nop                     ;
           nop                     ;
           nop                     ;

; Display the main playing area with sideborders disabled
; (10 rows with 8 sprites in each)
           ldy  #210               ; Main playing area is 210 raster lines
ib3        ldx  vicidxw,y          ; Get index to VIC register
           lda  vicvalw,y          ; Get value
           dec  v+22               ; Disable the sideborder
           sta  v,x                ; Set register value
           ldx  vicidxr,y          ; Get index to VIC register
           lda  v,x                ; Get register value
           sta  vicvalr,y          ; Store it
           inc  v+22               ; Enable the sideborder again
           bit  0                  ; Timing
           dey                     ;
           bne  ib3                ; Loop until done

; Check and update hiscore
; y is assumed to be zero here
hi1        lda  score,y            ; Get score digit
           cmp  hiscore,y          ; Compare to hiscore digit
           bcc  ib4                ; Value is less, so quit
           bne  hi2                ; Value is greater, so update hiscore
           iny                     ; Value is equal, so check next digit
           cpy  #5                 ;
           bne  hi1                ; Loop for all digits
hi2        ldy  #5-1               ;
hi3        lda  score,y            ; Transfer current score
           sta  hiscore,y          ; to hiscore
           dey                     ;
           bpl  hi3                ;

; Write score, hiscore, number of mines left etc. into sprite images
; (3 characters are written per frame)
ib4        ldx  txttabidx          ; Get current index
           dex                     ; Decrement it
           bpl  ib5                ; Do not reset yet
           ldx  #4                 ;
ib5        stx  txttabidx          ; Store new index value
           lda  txtsplo,x          ; Lo byte of address
           pha                     ;
           ldy  txtstart,x         ; Start index to text
           lda  txtend,x           ; End index to text
           sta  tmpa               ;
           lda  txtsphi,x          ; Hi byte of address
           tax                     ;
           pla                     ;
           jsr  writesptxt         ; Write characters

           lda  #249               ; Raster line for next interrupt
           jmp  ret                ;

; ******************************************************************************
; *  The third interrupt                                                       *
; ******************************************************************************

irqc       ldy  #7-1               ;
ic1        ldx  vicidxwbtm,y       ; Disable lower border and set up display of
           lda  vicvalwbtm,y       ; the bottom sprite row
           sta  v,x                ;
           dey                     ;
           bpl  ic1                ;
           ldx  #6-1               ;
           lda  #1                 ;
ic2        sta  v+39,x             ; Set white sprite text color
           dex                     ;
           bpl  ic2                ;
           lda  #lo irqd           ;
           sta  irqvector          ; Set lo byte of next interrupt routine
           lda  #hi irqd           ;
           sta  irqvector+1        ; Set hi byte of next interrupt routine

           dec  framecnt           ; Decrement frame counter
           bit  freezedsp          ; Freeze display?
           bvc  ic5                ; No
           bpl  ic3                ; Yes, but no need to acknowledge it
           lsr  freezedsp          ; Acknowledge freeze (shift bit 7 into bit 6)
ic3        ldx  #36                ;
ic4        dex                     ; Spend some raster time
           bpl  ic4                ;
           jmp  endirqc            ; Do not move subs, torpedoes etc.

ic5        lda  framecnt           ; Get frame counter value
           bit  gstatus            ; Is a game in progress?
           bvs  playerctrl         ; Yes, do game processing
           bpl  pc2                ; Title screen is not being displayed

; Blinking of "PRESS FIRE TO PLAY" text
bl1        tax                     ; Transfer frame counter to x
           and  #%00001111         ; Update text visibility now?
           bne  bl5                ; No
           clc                     ;
           txa                     ; Transfer frame counter to acc. again
           and  #%00111111         ; Make text invisible now?
           bne  bl2                ; No
           sec                     ; Yes, set carry flag to indicate it
           lda  #44                ; Blank sprite image
bl2        ldx  #85                ; Sprite image with "LAY"
           ldy  #6-1               ; Six sprite pointers to change
bl3        bcs  bl4                ; Make text invisible (blank sprite image)
           txa                     ; Text sprite image
bl4        sta  $67f8,y            ; Set sprite pointer
           dex                     ; Decrement sprite image number
           dey                     ; Decrement counter
           bpl  bl3                ; Loop until done
bl5        jmp  chksubrel          ;

playerctrl lsr  a                  ; Move mines and ship this frame?
           bcs  pc1                ; No
           inc  mineupdcnt         ; Initiate movement of mines
pc1        ldx  objexpcnt+18       ; Is the ship on fire?
           bmi  pc3                ; No
pc2        jmp  submove            ; Skip some game processing
pc3        bcs  checkfire          ; Do not move ship this frame

; Check joystick (move ship left or right)
           ldy  objxpos+18         ; Get lower 8 bits of x-coordinate
           lda  objxposmsb+18      ; Get MSB of x-coordinate (bit 0)
           lsr  a                  ; Shift MSB into carry flag
           lda  joystick           ; Get current joystick status
           tax                     ; Store in x for later use
           and  #%00000100         ; Move left?
           bne  checkright         ; No
           bcs  ml1                ; Ship is not in the left part of the screen
           cpy  #40                ; Has the ship reached the left edge?
           beq  checkfire          ; Yes, do not move it
ml1        ldx  #1                 ; Index to table of ship sprite images
           dey                     ; Decrement x-coordinate
           cpy  #255               ; Need to switch off MSB of x-coordinate?
           bne  mr2                ; No
           dec  objxposmsb+18      ; Yes
           beq  mr2                ; Branch always
checkright txa                     ; Get joystick status
           and  #%00001000         ; Move right?
           bne  checkfire          ; No
           bcc  mr1                ; Ship is not in the right part of the screen
           cpy  #136               ; Has the ship reached the right edge?
           beq  checkfire          ; Yes, do not move it
mr1        ldx  #0                 ; Index to table of ship sprite images
           iny                     ; Increment x-coordinate
           bne  mr2                ; No overflow
           inc  objxposmsb+18      ; Overflow - switch on MSB of x-coordinate
mr2        stx  fireimgidx+1       ; Self-modifying code
           lda  shipspimgl,x       ; Get sprite image
           sta  $53f8+4            ; Set sprite pointer for left part of ship
           lda  shipspimgr,x       ; Get sprite image
           sta  $53f8+5            ; Set sprite pointer for right part of ship
           tya                     ;
           sta  objxpos+18         ; Store x-coordinate of left ship sprite
           ldx  #18                ;
           jsr  dspobject          ; Calculate coordinates and display the ship

; Check fire button (drop mine)
checkfire  bit  endoflevel         ; Is the end-of-level indicator set?
           bmi  chksubrel          ; Yes, do not allow player to drop mines
           ldx  dropdelay          ; Get delay since last mine drop
           bmi  cf1                ; Delay countdown complete
           dec  dropdelay          ; Decrement delay counter
cf1        jsr  firebutton         ; Fire button pressed?
           beq  chksubrel          ; No
; A new mine can't be dropped if the delay countdown is in the range 0-5.
; Reason: Within this range, there's a risk that a mine sprite in the next row
; will be moved to a new x-coordinate while there's still a mine in it.
           txa                     ; Transfer counter to acc. to affect N flag
           bmi  cf3                ; Delay countdown complete - OK to drop mine
           cmp  #6                 ; Is the delay countdown in the unsafe range?
           bcc  chksubrel          ; Yes, do not drop a new mine now
cf3        lda  #45                ;
           sta  dropdelay          ; Initialize new drop delay
           jsr  dropmine           ; Drop a mine

chksubrel  ldy  #0                 ;
           lda  subrelcnt          ; Has the sub release counter reached zero?
           bne  checkmode          ; No
rs1        jsr  getdata            ; Get data byte
           tax                     ; Save in x for later use
           bmi  nextlevel          ; Bit 7 set indicates end of current level
           and  #%00000011         ; Isolate mode bits
           sta  subrelmode         ; Store sub release mode
           beq  rs2                ; Fixed mode, do not get sub rel interval now
           jsr  getdata            ; Get sub release interval
           sta  subrelval          ; Store in value variable
rs2        txa                     ; Get data byte again
           lsr  a                  ; Shift out mode bits
           lsr  a                  ;
           sta  subrelcnt          ; Store release counter
           lda  #255               ;
           sta  torpedosub         ; Turn off any "torpedo sub"
           bne  checkmode          ; Branch always

; x is assumed to be 255 here
nextlevel  bit  gstatus            ; Is a game in progress?
           bvc  rs3                ; No, restart demo
           stx  endoflevel         ; Set end-of-level indicator
           jsr  inclevel           ; Increment level number
           lda  #6                 ; Index to write level number into sprites
           sta  txttabidx          ; Store it
           sta  subrelval          ; Make sure sub release interval is non-zero
           sta  subrelint          ;
           bne  rs1                ; Branch always
rs3        lda  #lo demo           ; Lo byte of address
           sta  leveldata          ;
           lda  #hi demo           ; Hi byte of address
           sta  leveldata+1        ;
           bne  rs1                ; Branch always

checkmode  bit  endoflevel         ; Is the end-of-level indicator set?
           bpl  cm1                ; No
; Wait for all mines to become inactive before displaying "LEVEL n" message
           lda  minetabidx         ;
           bpl  submove            ;
           ldx  #105               ;
           stx  msgcounter         ; Counter for display of "LEVEL n" message
           jsr  setmsgspr          ; Set up message sprites
           jsr  resetmines         ; Reset number of mines
           inc  endoflevel         ; Reset end-of-level indicator

cm1        dec  subrelint          ; Decrement release interval counter
           bne  submove            ; Do not release a sub now
           dec  subrelcnt          ; Decrement release counter
           lda  subrelmode         ; Get current sub release mode
           beq  cm4                ; Fixed mode
           lda  subrelval          ; Get interval value
           sta  subrelint          ; Reset interval counter
           beq  submove            ; Interval is zero - do not release subs
           lda  s+27               ; Get a random value from SID oscillator 3
           sta  tmpa               ; Store direction (bit 7) in tmpa
           and  #31                ; Reduce value to range 0-31
           cmp  #18                ; Is it greater than 18?
           bcc  rn1                ; No
           sbc  #14                ; Yes, make sure value is in range 0-17
rn1        tax                     ;
           lda  subrelmode         ; Get current sub release mode
           lsr  a                  ; Shift 2nd mode bit into carry flag
           bcc  cm5                ; Two subs per sprite row
           lsr  a                  ; Shift 1st mode bit into carry flag
           txa                     ; Transfer random number (0-17) to acc.
           and  #%11111110         ; Only even sub numbers (sprites #4 and #5)
           bcc  cm2                ;
           ora  #%00000001         ; Only odd sub numbers (sprites #6 and #7)
cm2        tax                     ;
           bpl  cm5                ; Branch always

; Fixed mode: "Fast sub" is implied
cm4        jsr  getdata            ; Get sub info byte
           lsr  a                  ; Shift direction bit into carry flag
           ror  tmpa               ; Rotate into bit 7 of tmpa
           lsr  a                  ; Shift torpedo bit into carry flag
           tax                     ; Submarine index
           stx  fastsub            ; Store index of "fast sub"
           bcc  cm6                ; This sub should not launch a torpedo
           stx  torpedosub         ; Store index of "torpedo sub"
cm6        jsr  getdata            ; Get sub release interval
           sta  subrelval          ; Store in value variable
           sta  subrelint          ; Store in counter variable

; x is assumed to contain submarine index here
cm5        lda  objxposmsb,x       ;
           bpl  submove            ; Do not reset an active sub
           lda  objexpcnt,x        ;
           bpl  submove            ; Or a sub that's currently exploding
           jsr  releasesub         ; Release the submarine

; Move 9 submarines every odd/even frame (+ any "fast sub")
submove    ldy  #18                ; Default: Do not activate a "torpedo sub"
           lda  subrelmode         ;
           cmp  #1                 ; Is the game in torpedo launch mode?
           bne  sm1                ; No
           ldy  #2                 ; Yes, a "torpedo sub" may be activated
sm1        lda  framecnt           ; Get frame counter value
           lsr  a                  ; Determine which subs to move
           ldx  #0                 ; Start index
           lda  #9                 ; End index + 1
           bcc  sm2                ;
           tax                     ; Start index
           lda  #18                ; End index + 1
sm2        jsr  movesubs           ; Move submarines
           ldx  fastsub            ; Is a "fast sub" active?
           bmi  tp1                ; No
           jsr  chksubmove         ; Yes, check if it needs to be moved now
           bcs  tp1                ; No, it has already been moved this frame
           txa                     ;
           adc  #1                 ; End index + 1
           ldy  #18                ;
           jsr  movesubs           ; Move the "fast sub"

; Torpedo launch/movement
tp1        lda  torpedocnt         ; Is a torpedo active?
           beq  tp2                ; No
           jsr  movetorp           ; Yes, move it
           jmp  tp3                ;
tp2        ldx  torpedosub         ; Is a "torpedo sub" active?
           bmi  tp3                ; No
           lda  objexpcnt+18       ; Is the ship already on fire?
           bpl  tp3                ; Yes, do not attempt to launch a torpedo
           jsr  launchtorp         ; Check range, launch torpedo

; Check game status
tp3        bit  gstatus            ; Is a game in progress?
           bvc  go3                ; No
           ldx  msgcounter         ; Is "LEVEL n" message being displayed?
           bmi  tp4                ; No
           dex                     ; Yes, decrement counter
           stx  msgcounter         ; Store new value
           bne  tp4                ; It's not zero yet
           lda  #44                ;
           sta  $67f8              ; Replace message with blank sprites
           sta  $67f9              ; and a mine sprite image
           lda  #5                 ;
           sta  $67fa              ;
           sec                     ;
           ldy  #3                 ;
           jsr  uv1                ; Restore black color for mine sprites
tp4        lda  isshiplost         ; Has a ship just been lost?
           beq  go2                ; No
           ldx  ships              ; Are there any ships left?
           bpl  go7                ; Yes, continue game
; Wait for all mines to become inactive before displaying "GAME OVER" message
           lda  minetabidx         ;
           bpl  go2                ;
           ldx  #102               ;
           jsr  setmsgspr          ; Set up "GAME OVER" message
           lsr  gstatus            ; Change status to "Game over"
           bne  go3                ; Branch always

go7        jsr  setupship          ; Set up ship's x-coord. and sprite images
           dec  isshiplost         ; Clear "ship lost" flag
           ldx  #255               ;
           bne  go6                ; Stop "ship on fire" animation

go2        jsr  checkcoll          ; Check for collisions

; "Ship on fire" animation
go3        ldx  objexpcnt+18       ; Is the ship on fire?
           bmi  endirqc            ; No
           lda  framecnt           ; Yes
           lsr  a                  ; Update animation counter now?
           bcc  endirqc            ; No
           txa                     ; Yes, transfer counter to acc.
           lsr  a                  ;
           lsr  a                  ;
           lsr  a                  ;
fireimgidx ldy  #0                 ; Determine which animation frame
           bcc  go4                ; to display
           iny                     ;
           iny                     ;
go4        lda  shipspimgl+2,y     ; Get sprite image
           sta  $53f8+4            ; Set sprite pointer for left part of ship
           lda  shipspimgr+2,y     ; Get sprite image
           sta  $53f8+5            ; Set sprite pointer for right part of ship
           dex                     ; Decrement animation counter
           bne  go6                ; Countdown not complete yet
           lda  isshiplost         ; Is this the first completed countdown?
           bne  go5                ; No
           inc  isshiplost         ; Yes, set "ship lost" flag to indicate it
           dec  ships              ; and decrement number of ships left
go5        ldx  #96                ; Reset animation counter
go6        stx  objexpcnt+18       ; Store new counter value

endirqc    lda  #23                ;
ic6        cmp  v+18               ; Need to wait until line 24 to avoid double
           bcs  ic6                ; display of the top sprite row

           ldx  #6-1               ;
ic7        lda  vicvalwval,x       ;
           sta  v+39,x             ; Set sprite colors
           dex                     ;
           bpl  ic7                ;

           ldy  #19-1              ;
ic8        ldx  vicidxwtop,y       ; Set up display of top sprite row
           lda  vicvalwtop,y       ;
           sta  v,x                ;
           dey                     ;
           bpl  ic8                ;

           lda  #14                ; Raster line for next interrupt
           bne  ret                ; Branch always

; ******************************************************************************
; *  The fourth interrupt                                                      *
; ******************************************************************************

irqd       jsr  setirqa            ; Set irq vector for the first interrupt

           lda  #14                ; Color of sky (light blue)
           sta  v+32               ; Set border color
           sta  v+33               ; Set background color

           lda  #21                ; Raster line for next interrupt
ret        sta  v+18               ;
           dec  v+25               ; Acknowledge current interrupt
           jmp  $ea81              ; Return from interrupt

; ******************************************************************************
; *  Release a submarine                                                       *
; *  Inputs:  x       Submarine index                                          *
; *           tmpa    Bit 7 is direction (0=move right)                        *
; ******************************************************************************

releasesub stx  tmpc               ; Save index of sub to be released
           txa                     ;
           lsr  a                  ; Is the sub index even?
           bcc  su5                ; Yes
           dex                     ; Calculate index of even sub on same row
           db   $24                ; Skip next byte using a bit instruction
su5        inx                     ; Calculate index of odd sub on same row
           lda  objxposmsb,x       ; Is the other sub active?
           bmi  su6                ; No (bit 7 set)
; Make sure that two subs in the same row are not moving in opposite directions
           asl  a                  ; Shift direction bit into bit 7
           eor  tmpa               ; Same direction for both subs in the row?
           asl  a                  ;
           bcc  su6                ; Yes
           lda  tmpa               ; No, need to change direction of
           eor  #%10000000         ; the sub to be released
           sta  tmpa               ;
su6        ldx  tmpc               ; Get index of sub to be released
           lda  #0                 ; Initial x-coordinate
           tay                     ; Initial x-coordinate MSB
           asl  tmpa               ; Move sub right?
           php                     ;
           bcc  su1                ; Yes
           lda  #192               ; Initial x-coordinate
           ldy  #%01000001         ; Initial x-coordinate MSB + left-flag
su1        sta  objxpos,x          ; Virtual x-coordinate
           sty  objxposmsb,x       ; Activate sub
           jsr  chksubmove         ; Will the sub be moved this frame?
           bcs  su2                ; Yes
           lda  objxpos,x          ;
           jsr  dspobject          ; No, it needs to be displayed now
su2        jsr  calcsprptr         ; Calculate sprite pointer address
           ldy  #0                 ;
           lda  #36                ; Left sprite image of sub moving right
           plp                     ; Move sub right?
           bcc  su3                ; Yes
           lda  #53                ; Left sprite image of sub moving left
su3        sta  (va),y             ; Set sprite pointer
           iny                     ;
           lda  #37                ; Right sprite image of sub moving right
           bcc  su4                ;
           lda  #52                ; Right sprite image of sub moving left
su4        sta  (va),y             ; Set sprite pointer
           rts

; ******************************************************************************
; *  Move/explode submarines                                                   *
; *  Inputs:  a       End index + 1                                            *
; *           x       Start index                                              *
; *           y       Index limit for activating a "torpedo sub"               *
; ******************************************************************************

movesubs   sty  ms10+1             ; Self-modifying code
           sta  tmpb               ; Store end index
ms0        lda  objxposmsb,x       ; Get MSB (bit 0) and status (bit 6 and 7)
           bmi  ms6                ; Sub is currently disabled (bit 7 set)
           ldy  objxpos,x          ; Get x-coordinate of left sub sprite
           lsr  a                  ; Shift MSB of x-coordinate into carry flag
           and  #%00100000         ; Move sub left?
           bne  ms3                ; Yes
           bcc  ms1                ; Sub is not in the right part of the screen
           cpy  #192               ; Right edge reached? (448-256=192)
           bne  ms1                ; No
ms8        ror  objxposmsb,x       ; Disable the sub
           lda  #255               ;
           cpx  fastsub            ; Was this a "fast sub"?
           bne  ms9                ; No
           sta  fastsub            ; Yes, turn it off
ms9        cpx  torpedosub         ; Was this a "torpedo sub"?
           bne  ms6                ; No
           sta  torpedosub         ; Yes, turn it off
           beq  ms6                ; Branch always
ms3        bcs  ms7                ; Sub is not in the left part of the screen
           cpy  #0                 ; Left edge reached?
           beq  ms8                ; Yes
ms7        dey                     ; Decrement x-coordinate
           cpy  #255               ; Need to switch off MSB of x-coordinate?
           bne  ms2                ; No
           dec  objxposmsb,x       ; Yes
           bne  ms2                ; Branch always
ms1        iny                     ; Increment x-coordinate
           bne  ms2                ; No overflow
           inc  objxposmsb,x       ; Overflow - MSB of x-coordinate is set
ms2        tya                     ;
           sta  objxpos,x          ; Store x-coordinate of left sub sprite
           jsr  dspobject          ; Calculate coordinates and display the sub
           lda  objexpcnt,x        ; Is the sub currently exploding?
           bpl  ms4                ; Yes
ms10       cpx  #0                 ; Check index limit
           bcc  ms6                ; Do not activate a "torpedo sub"
           lda  torpedosub         ; Is such a sub already active?
           bpl  ms6                ; Yes
           stx  torpedosub         ; No, make current sub a "torpedo sub"
           bmi  ms6                ; Branch always
; Explosions
ms4        lsr  a                  ; Update the explosion frame now?
           bcs  ms5                ; No
           jsr  calcsprptr         ; Calculate sprite pointer address
           ldy  #0                 ;
           lda  (va),y             ; Get current frame for left part
           adc  #2                 ; Increment frame
           sta  (va),y             ; Set sprite pointer for left part of sub
           iny                     ;
           lda  (va),y             ; Get current frame for right part
           adc  #2                 ; Increment frame
           sta  (va),y             ; Set sprite pointer for right part of sub
ms5        dec  objexpcnt,x        ; Decrement explosion counter
           bpl  ms6                ; Explosion not finished yet
           lda  #%10000000         ;
           sta  objxposmsb,x       ; Disable the sub
ms6        inx                     ; Next sub
           cpx  tmpb               ;
           bne  ms0                ; Loop until done
           rts                     ;

; ******************************************************************************
; *  Check if a submarine is moved this frame (odd/even)                       *
; *  Inputs:  x       Submarine index                                          *
; *  Outputs: C       Set if the specified sub is moved this frame             *
; ******************************************************************************

chksubmove cpx  #9                 ; Is the specified sub in the range 9-17?
           lda  framecnt           ; Get frame counter value
           and  #1                 ; Check if odd or even
           beq  cd1                ; Branch if subs 0-8 are moved this frame
           rts                     ; Return carry flag from compare operation
cd1        rol  a                  ;
           eor  #1                 ; Invert carry flag from compare operation
           lsr  a                  ;
           rts

; ******************************************************************************
; *  Calculate sprite pointer address for a submarine                          *
; *  Inputs:  x       Submarine index                                          *
; *  Outputs: va      Sprite pointer address                                   *
; ******************************************************************************

calcsprptr txa                     ; Transfer submarine index to acc.
           ldy  #252               ; Lo byte of sprite pointer address
           lsr  a                  ; Shift out bit 0
           bcc  cs1                ; Submarine index is even
           ldy  #254               ; Lo byte of sprite pointer address
cs1        sty  va                 ; Store lo byte
           asl  a                  ; Calculate hi byte of sprite pointer address
           asl  a                  ;
           adc  #87                ;
           sta  va+1               ; Store hi byte
           rts

; ******************************************************************************
; *  Calculate actual x-coordinates and set up display of object (2 sprites)   *
; *  Inputs:  a       Virtual x-coordinate (lower 8 bits)                      *
; *           x       Index to tables 'objxpos', 'objxposmsb' and 'valtabidx'  *
; ******************************************************************************

dspobject  clc                     ;
           adc  #24                ; Calculate virtual x-coord. of right sprite
           bcs  do1                ; Overflow - MSB is set for the right sprite
           tay                     ;
           lda  objxposmsb,x       ; No overflow - use the same MSB for the
           lsr  a                  ; left and right sprite
           tya                     ;
; Set up display of the right sprite
do1        jsr  getactualx         ; Calculate actual VIC x-coordinate
           ldy  valtabidx,x        ; Get index to VIC register values table
           sta  vicvalw,y          ; Store actual x-coordinate of right sprite
           php                     ; Save carry flag (MSB of x-coordinate)
; Set up display of the left sprite
           lda  objxposmsb,x       ; Get MSB of virtual x-coordinate
           lsr  a                  ; Shift MSB into carry flag
           lda  objxpos,x          ; Get virtual x-coordinate of left sprite
           jsr  getactualx         ; Calculate actual VIC x-coordinate
           sta  vicvalw+1,y        ; Store actual x-coordinate of left sprite
           php                     ; Save carry flag (MSB of x-coordinate)
; Set up value for v+16 according to MSB values
           txa                     ; Transfer table index to acc.
           stx  tmpa               ; Also store it in a temporary variable
           ldx  #2                 ; Index to bit table for sprites 6/7
           lsr  a                  ; Identify which sprites to use
           bcs  do2                ; Use sprites 6/7
           ldx  #0                 ; Index to bit table for sprites 4/5
           dey                     ; When using sprites 4/5, adjust index to
           dey                     ; VIC register values table accordingly
do2        lda  vicvalw-1,y        ; Current v+16 value for this sprite row
           and  andmask3,x         ; Mask bits according to sprite #
           plp                     ; Pull MSB of left sprite from stack
           bcc  do3                ; Not set
           ora  oramask1+4,x       ; Set the appropriate v+16 bit
do3        plp                     ; Pull MSB of right sprite from stack
           bcc  do4                ; Not set
           ora  oramask1+5,x       ; Set the appropriate v+16 bit
do4        sta  vicvalw-1,y        ; Store new v+16 value
           ldx  tmpa               ; Restore table index
           rts

; ******************************************************************************
; *  Convert from virtual x-coordinate to actual VIC x-coordinate              *
; *  Inputs:  a       Virtual x-coordinate (lower 8 bits)                      *
; *           C       Virtual x-coordinate MSB                                 *
; *  Outputs: a       Actual x-coordinate (lower 8 bits)                       *
; *           C       Actual x-coordinate MSB                                  *
; ******************************************************************************

; The game uses x-coordinates 0-448 internally. Because sprites are also
; displayed in the sideborders, the 'virtual' coordinates are mapped to
; actual VIC x-coordinates as follows:
;
; Virtual    MSB      Actual    MSB
; 0-63        0       184-247*   1
; 64-255      0      *0-191      0
; 0-63        1       192-255    0
; 64-192      1       0-128      1
;
; (* There's a 'gap' at VIC coordinates 248-255 where sprites are not displayed)

getactualx bcs  ga1                ; Virtual x-coordinate > 255
           cmp  #64                ;
           bcs  ga0                ;
           adc  #184               ;
           sec                     ; MSB set
           rts                     ;
ga0        sbc  #64                ;
           clc                     ; MSB clear
           rts                     ;
ga1        cmp  #64                ;
           bcs  ga2                ;
           adc  #192               ;
           rts                     ;
ga2        sbc  #64                ;
ga3        rts

; ******************************************************************************
; *  Check and launch torpedo from sub                                         *
; ******************************************************************************

launchtorp lda  objxpos,x          ; X-coord. of left sub sprite
           sec                     ;
           sbc  objxpos+18         ; Subtract x-coord. of left ship sprite
           sta  va                 ; Save result
           lda  objxposmsb,x       ;
           and  #%00111111         ; MSB of left sub sprite's x-coord.
           sbc  objxposmsb+18      ; Subtract MSB of left ship sprite
           bcs  lt1                ; Branch if sub's x-coord. > ship's x-coord.
           tay                     ;
           lda  va                 ; Calculate absolute value
           eor  #255               ;
           adc  #1                 ;
           sta  va                 ;
           tya                     ;
           eor  #255               ;
           adc  #0                 ;
lt1        bne  ga3                ; Diff > 255: No torpedo launch
           lda  va                 ;
           cmp  #25                ;
           bcs  ga3                ; Diff > 24: No torpedo launch
; Set up sprite pointers for torpedo movement
           lda  #$fe               ; Lo byte of sprite pointer address
           sta  va                 ;
           lda  #$53               ; Hi byte of sprite pointer address
           sta  va+1               ;
           lda  #10-1              ; 10 sprite rows to process
           sta  tmpa               ;
           ldx  #20                ; First sprite image for torpedo
           ldy  #0                 ;
lt2        txa                     ;
           sta  (va),y             ; Set sprite pointer
           lda  va+1               ;
           adc  #4                 ; Adjust sprite pointer address
           sta  va+1               ;
           inx                     ; Increment sprite image
           dec  tmpa               ;
           bpl  lt2                ; Loop until done
; Set up x-coordinates of sprites for torpedo movement
           ldx  torpedosub         ;
           ldy  #21                ; Byte offset of torpedo (sub moving left)
           lda  objxposmsb,x       ; Get MSB (bit 0) and direction (bit 6)
           lsr  a                  ; Shift MSB into carry flag
           and  #%00100000         ; Is the sub moving left?
           bne  lt3                ; Yes
           ldy  #23                ; Byte offset of torpedo (sub moving right)
           lda  objxpos,x          ; Left sub sprite's x-coord.
           clc                     ;
           adc  #24                ; Calculate x-coord. of torpedo sprite
           bcs  lt4                ; Overflow - MSB is set
           pha                     ;
           lda  objxposmsb,x       ; No overflow - use the same MSB
           lsr  a                  ;
           pla                     ;
           db   $2c                ; Skip next two bytes using a bit instruction
lt3        lda  objxpos,x          ; Left sub sprite's x-coord.
lt4        sty  torpspoffs         ; Store byte offset of torpedo
           jsr  getactualx         ; Calculate actual VIC x-coordinate
           ror  tmpa               ; Rotate actual MSB into bit 7 of tmpa
           tay                     ; Transfer actual x-coord. to y
           ldx  #210               ; 10 sprite rows to process (10 x 21)
           sec                     ;
lt5        tya                     ; Get actual x-coord.
           sta  vicvalw+1+3,x      ; Store it
           lda  vicvalw+1+1,x      ; Current v+16 value for this sprite row
           and  #%10111111         ; Mask bits
           bit  tmpa               ; Should MSB be set for torpedo sprites?
           bpl  lt6                ; No (bit 7 not set)
           ora  #%01000000         ; Yes
lt6        sta  vicvalw+1+1,x      ; Store new v+16 value
           txa                     ;
           sbc  #21                ; Adjust index to 'vicvalw' table
           tax                     ;
           bne  lt5                ; Loop until done
; Calculate sprite image address and torpedo's distance to water surface
           lda  #0                 ;
           sta  torpspimg+1        ;
           lda  torpedosub         ; Submarine index
           asl  a                  ; Multiply by 2
           sta  tmpb               ; Save for later use (distance calculation)
           asl  a                  ; Calculate address of sprite image
           asl  a                  ;
           asl  a                  ;
           rol  torpspimg+1        ;
           asl  a                  ;
           rol  torpspimg+1        ;
           adc  #$40               ;
           sta  torpspimg          ;
           lda  torpspimg+1        ;
           adc  #$45               ;
           sta  torpspimg+1        ;
           ldx  #%11000000         ; Torpedo bit pattern for sub moving left
           lda  torpspoffs         ;
           cmp  #21                ; Is the sub moving left?
           beq  lt7                ; Yes
           ldx  #%00000011         ; Torpedo bit pattern for sub moving right
lt7        stx  movetorp+1         ; Store bit pattern (self-modifying code)
           lda  torpedosub         ;
           lsr  a                  ;
           sta  tmpa               ; Number of sprite rows above sub
           lda  tmpb               ;
           asl  a                  ; Calculate distance from sub launching
           asl  a                  ; torpedo to water surface
           adc  tmpa               ; (number of sprite rows * 21 + 16 pixels)
           adc  tmpb               ;
           adc  #16                ;
           sta  torpedocnt         ; Store in movement counter
           rts

; ******************************************************************************
; *  Move torpedo                                                              *
; ******************************************************************************

movetorp   lda  #0                 ; Bit pattern for drawing torpedo in sprite
           dec  torpedocnt         ; Decrement movement counter
           bne  mt1                ; It's not zero yet
           lda  #255               ; Torpedo has reached the water surface
           sta  torpedosub         ; A new sub may now launch a torpedo
           lda  #%00000000         ; Bit pattern for erasing existing torpedo
mt1        ldx  #4-1               ; Height of torpedo = 4 pixels
           ldy  torpspoffs         ; Get byte offset within sprite
mt2        sta  (torpspimg),y      ; Draw torpedo
           iny                     ; Adjust index
           iny                     ;
           iny                     ;
           cpy  #63                ; Need to adjust index further if the torpedo
           beq  mt3                ; has reached a new sprite row (image)
           cpy  #65                ;
           bne  mt4                ;
mt3        iny                     ;
mt4        dex                     ;
           bpl  mt2                ; Loop until entire torpedo has been drawn
           lda  #%00000000         ;
           sta  (torpspimg),y      ; Erase bottom of torpedo
           lda  torpspoffs         ;
           sec                     ;
           sbc  #3                 ; Adjust index
           bcs  mt6                ; Still in the same sprite row (image)
           lda  torpspimg          ;
           sec                     ; Adjust sprite image address
           sbc  #64                ;
           sta  torpspimg          ;
           bcs  mt5                ;
           dec  torpspimg+1        ;
mt5        lda  torpspoffs         ;
           clc                     ;
           adc  #60                ; Reset byte offset within sprite
mt6        sta  torpspoffs         ;
           rts

; ******************************************************************************
; *  Check for collisions (mines/subs, torpedo/ship)                           *
; ******************************************************************************

; Check for collisions between mines and subs
checkcoll  lda  #0                 ;
           sta  disablecnt         ; Reset number of mines disabled this frame
           ldx  #1                 ; Sprite row number (1-9)
           ldy  #170               ; Index to 'vicvalr' table
cc1        sty  tmpa               ; Save index in a temporary variable
           lda  vicvalr,y          ; Get sprite collision status for this row
           tay                     ; Save in y register
           and  #%00001111         ; Is any mine sprite involved in collision?
           beq  cc3                ; No
           stx  tmpb               ; Yes, save current sprite row number
           dex                     ; Calculate submarine index
           txa                     ;
           asl  a                  ;
           tax                     ;
           tya                     ; Get sprite collision status
           and  #%11000000         ; Is sprite #6 or #7 involved?
           beq  cc2                ; No
           inx                     ; Yes, adjust submarine index
cc2        lda  objxposmsb,x       ; Is the submarine disabled?
           bmi  cc7                ; Yes
           lda  objexpcnt,x        ; Is the submarine already exploding?
           bpl  cc7                ; Yes
           lda  #128               ;
           sta  s+4                ; Terminate previous sub explosion sound
           inc  s+4                ; Initiate new sub explosion sound
           lda  #6                 ; Initiate explosion animation
           sta  objexpcnt,x        ;
           cpx  fastsub            ; Is this a "fast sub"?
           bne  cc12               ; No
           lda  #255               ; Yes, explosion should run at normal
           sta  fastsub            ; speed and 1000 pts should be awarded
cc12       cpx  torpedosub         ; Is this a "torpedo sub"?
           bne  cc10               ; No
           ror  torpedosub         ; Yes, turn it off
cc10       jsr  addpts             ; Add points to score
; Disable mines involved in the collision
           ldx  tmpb               ; Get current sprite row number
           tya                     ; Get sprite collision status
           and  #%00001010         ; Is sprite #1 or #3 involved?
           beq  cc4                ; No (mine sprite # = current sprite row #)
           txa                     ;
           clc                     ; Yes, calculate mine sprite number
           adc  #10                ;
           tax                     ;
cc4        stx  tmpc               ; Mine sprite number for current row
           dex                     ;
           stx  tmpd               ; Mine sprite number for previous row
           ldx  #10-1              ; Max 10 mines moving concurrently
cc5        lda  minespnum,x        ; Get sprite number for current mine
           cmp  tmpc               ; Is this mine located in the current sprite?
           beq  cc8                ; Yes
           cmp  tmpd               ; No, is it located in the previous sprite?
           bne  cc6                ; No
           ldy  mineypos,x         ; Get mine's y-coordinate within sprite
           cpy  #19                ; Mine present in both prev. and curr. row?
           bcc  cc6                ; No
cc8        ror  minespnum,x        ; Disable the mine
           stx  minenum            ; Save mine index
           tax                     ; Transfer mine sprite number to x
           dec  mspminecnt,x       ; Decrement number of mines in sprite
           ldx  minenum            ; Restore mine index
           jsr  erasemine          ; Erase the mine from the screen
           inc  disablecnt         ; Inc number of mines disabled this frame
cc6        dex                     ; Next mine
           bpl  cc5                ; Loop until done
           lda  disablecnt         ; Get number of mines disabled this frame
           cmp  #5                 ; Is it 5 or more?
           bcs  cc9                ; Yes, terminate collision check (mines/subs)
cc7        ldx  tmpb               ; Get current sprite row number
cc3        inx                     ; Increment current sprite row number
           lda  tmpa               ; Get current index to 'vicvalr' table
           sec                     ;
           sbc  #21                ; Adjust for next sprite row
           tay                     ;
           bcs  cc1                ; Loop until done

; Check for collisions between torpedo and ship
cc9        lda  objexpcnt+18       ; Is the ship already on fire?
           bpl  cc11               ; Yes, do not check for collision
           lda  #%00110000         ; Bit mask for ship sprites
           bit  vicvalr+191        ; Check sprite collision status for top row
           bvc  cc11               ; Torpedo is not involved in any collision
           beq  cc11               ; Ship is not involved in any collision
           lda  #96                ;
           sta  objexpcnt+18       ; Set the ship on fire
           lda  #128               ;
           sta  s+11               ; Terminate previous ship explosion sound
           inc  s+11               ; Initiate new ship explosion sound
cc11       rts

; ******************************************************************************
; *  Drop a new mine from the ship                                             *
; ******************************************************************************

; Find a free mine (a currently disabled mine)
dropmine   ldx  #10-1              ; Max 10 mines moving concurrently
dm1        lda  minespnum,x        ; Get mine sprite number (0-19)
           bmi  dm3                ; This mine is free
           dex                     ; Next mine
           bpl  dm1                ; Loop until done
dm2        rts                     ; No free mines or no mines left
; Decrement number of mines left
dm3        lda  mines+2            ; Get 2nd digit
           cmp  #7                 ; Is it a zero?
           bne  dm4                ; No
           lda  mines+1            ; Get 1st digit
           cmp  #7                 ; Is it a zero?
           beq  dm2                ; Yes, there are no mines left
           sbc  #8                 ; Decrement 1st digit
           sta  mines+1            ; Store it
           lda  #9*8+7             ; Reset 2nd digit
           db   $2c                ; Skip next two bytes using a bit instruction
dm4        sbc  #8                 ; Decrement 2nd digit
           sta  mines+2            ; Store it
; Determine which sprite to use for the mine
           stx  minenum            ; Store index to the free mine
           lda  #23                ; Default: New mine sprite at ship+23
           sta  tmpa               ;
           clc                     ;
           adc  objxpos+18         ;
           sta  va                 ;
           lda  objxposmsb+18      ;
           adc  #0                 ;
           sta  va+1               ;
           ldx  #0                 ; Index to 'mspxpos' for the 1st mine sprite
           lda  mspminecnt         ; Are there any mines in the 1st sprite?
           beq  dm5                ; No
           ldy  #0                 ; Index to 'mspxpos' for mine sprite to check
           jsr  checkrange         ; Check if the 1st sprite is within range
           bcc  setupmine          ; Yes, reuse 1st sprite at current x-coord.
           ldx  #10                ; No, MUST reuse or change pos of 2nd sprite
dm5        lda  mspminecnt+10      ; Are there any mines in the 2nd sprite?
           beq  changepos          ; No
           ldy  #10                ; Index to 'mspxpos' for mine sprite to check
           jsr  checkrange         ; Check if the 2nd sprite is within range
           bcs  changepos          ; No, change position of existing sprite
           ldx  #10                ; Yes, reuse it

; Set up mine in sprite
setupmine  inc  mspminecnt,x       ; Increment number of mines within sprite
           txa                     ; Mine sprite number
           ldx  minenum            ; Get mine index
           sta  minespnum,x        ; Store mine sprite number
           lda  tmpa               ; X-coord. diff between mine sprite and ship
           cmp  #24                ;
           bcc  gm3                ; Make sure range is 0-23
           sbc  #24                ;
gm3        lsr  a                  ; Divide by 8
           lsr  a                  ;
           lsr  a                  ;
           eor  #%00000011         ; Calculate byte offset within sprite
           tay                     ;
           dey                     ;
           sty  minespoffs,x       ; Store byte offset
           lda  #0                 ;
           sta  mineypos,x         ; Reset y-coordinate within sprite
           lda  tmpa               ; X-coord. diff between mine sprite and ship
           and  #%00000111         ; Calculate index to bit pattern tables
           sta  minebitidx,x       ; Store it
           rts

; Change x-coordinate of mine sprite indicated by x-reg. (0=1st, 10=2nd)
changepos  stx  tmpb               ; Save mine sprite number
           lda  va+1               ; MSB of x-coordinate
           sta  mspxposmsb,x       ;
           lsr  a                  ; Shift into carry flag
           lda  va                 ; X-coordinate
           sta  mspxpos,x          ;
           jsr  getactualx         ; Calculate actual VIC x-coordinate
           ldy  valtabidx+19,x     ; Get index to VIC register values table
           sta  vicvalw,y          ; Store actual x-coordinate of mine sprite
           php                     ; Save carry flag (MSB of x-coordinate)
           cpx  #10                ; Is this the 2nd mine sprite?
           ldx  #0                 ;
           bcc  gm8                ; No
           inx                     ; Yes, adjust index to bit table
           iny                     ; Also adjust index to 'vicvalw' table
gm8        lda  vicvalw-8,y        ; Current v+16 value for the top sprite row
           and  andmask1,x         ; Mask bits according to sprite #
           plp                     ; Pull MSB of x-coordinate from stack
           bcc  gm5                ; MSB not set
           ora  oramask1,x         ; Set the appropriate v+16 bit
gm5        sta  vicvalw-8,y        ; Store new v+16 value
           lda  vicvalw-8-21,y     ; Current v+16 value for the next sprite row
           and  andmask1,x         ; Mask bits according to sprite #
           bcc  gm6                ; MSB not set
           ora  oramask1,x         ; Set the appropriate v+16 bit
gm6        sta  vicvalw-8-21,y     ; Store new v+16 value
           ldx  tmpb               ; Restore mine sprite number
           bpl  setupmine          ; Branch always

; ******************************************************************************
; *  Check if current mine sprite x-coordinate is within range for reuse       *
; *  Inputs:  y       Index to tables 'mspxpos' and 'mspxposmsb'               *
; *  Outputs: tmpa    X-coordinate diff between mine sprite and ship           *
; *           va      New sprite coordinate if reuse not possible              *
; *           C       Not set if sprite may be reused at current x-coordinate  *
; ******************************************************************************

checkrange lda  mspxpos,y          ; Mine sprite x-coordinate
           sec                     ;
           sbc  objxpos+18         ; Subtract x-coord. of left ship sprite
           sta  tmpa               ; Save result for later use
           lda  mspxposmsb,y       ; MSB of mine sprite x-coordinate
           sbc  objxposmsb+18      ; Subtract MSB of left ship sprite
           bcc  cr2                ; Diff < 0:   New sprite at ship+23
           bne  cr3                ; Diff > 255: New sprite at ship+23
           lda  tmpa               ;
           cmp  #24                ;
           bcc  cr1                ; Reuse existing sprite
           cmp  #48                ;
           bcs  cr3                ; Diff > 47:  New sprite at ship+23
           lda  mspxpos,y          ; 24 <= Diff <= 47
           sec                     ; New sprite at existing pos - 24
           sbc  #24                ;
           sta  va                 ;
           lda  mspxposmsb,y       ;
           sbc  #0                 ;
           sta  va+1               ;
cr1        rts                     ;
cr2        sec                     ;
cr3        lda  #23                ; New sprite at ship+23
           sta  tmpa               ;
           rts

; ******************************************************************************
; *  Move mines                                                                *
; ******************************************************************************

movemines  lda  #255               ;
           sta  minetabidx         ; Reset mine table index
           ldx  #10-1              ; Max 10 mines moving concurrently
mm1        ldy  minespnum,x        ; Get mine sprite number (0-19)
           bpl  mm12               ; Mine is active
mm3        dex                     ; Next mine
           bpl  mm1                ; Loop until done
           rts
mm12       stx  minetabidx         ; Save mine table index
           tya                     ; Transfer mine sprite number to acc.
           sta  islastrow          ; Range is 0-19, so bit 7 will be cleared
           cmp  #10                ; Determine sprite row number
           bcc  mm6                ;
           sbc  #10                ;
mm6        cmp  #9                 ; Is the mine in the last sprite row?
           bne  mm11               ; No
           ror  islastrow          ; Yes, set bit 7 to indicate it
mm11       lda  mspdatalo,y        ; Get lo byte of sprite image address
           sta  vb                 ;
           lda  mspdatahi,y        ; Get hi byte of sprite image address
           sta  vb+1               ;
           lda  mineypos,x         ; Get mine's y-coordinate within sprite
           asl  a                  ; Multiply by 2
           adc  mineypos,x         ; Value is now multiplied by 3
           adc  minespoffs,x       ; Add byte offset within sprite image
           tay                     ; Actual sprite image byte index
           lda  minebitidx,x       ; Get index to bit pattern tables
           tax                     ;
           lda  (vb),y             ;
           and  mineandval,x       ; Erase top of mine
           sta  (vb),y             ;
           ldx  minetabidx         ; Get mine table index
           lda  mineypos,x         ; Get mine's y-coordinate within sprite
           cmp  #18                ; Time to set up sprite in the next row?
           bne  mm4                ; No
           bit  islastrow          ; Yes, but is this the last sprite row?
           bpl  setnextrow         ; No, set up the next row
mm4        cmp  #20                ; Mine about to leave this sprite row?
           bne  mm5                ; No
           lda  minespnum,x        ; Yes
           tax                     ;
           dec  mspminecnt,x       ; Decrement number of mines in current sprite
           ldx  minetabidx         ; Get mine table index
           bit  islastrow          ; Is this the last sprite row?
           bpl  mm8                ; No
           ror  minespnum,x        ; Yes, disable the mine
           bne  mm3                ; Branch always
mm8        inc  minespnum,x        ; Transfer mine to next sprite row (image)
           lda  #255               ;
           sta  mineypos,x         ; Reset y-coordinate (increased by 1 below)
; Draw the mine
mm5        lda  minebitidx,x       ; Get index to bit pattern tables
           tax                     ;
           lda  #3-1               ; Height of mine = 3 pixels
           sta  tmpcnt             ;
mm2        iny                     ; Adjust index
           iny                     ;
           iny                     ;
           cpy  #63                ; Need to adjust index further if the mine
           bcc  mm7                ; has reached a new sprite row (image)
           cpy  #66                ;
           bcs  mm7                ;
           bit  islastrow          ; Is this the last sprite row?
           bmi  mm10               ; Yes, do not continue drawing this mine
           iny                     ; No, adjust index for next sprite image
mm7        lda  (vb),y             ;
           ora  mineoraval,x       ; Draw mine pixels
           sta  (vb),y             ;
           dec  tmpcnt             ;
           bpl  mm2                ; Loop until the entire mine has been drawn
mm10       ldx  minetabidx         ; Restore mine table index
           inc  mineypos,x         ; Increment y-coordinate within sprite
           jmp  mm3                ; Next mine

; Set x-coordinate for sprite in the next row
setnextrow sty  sn6+1              ; Save index (self-modifying code)
           lda  minespnum,x        ; Mine sprite number
           tax                     ;
           lsr  a                  ; Determine if row number is odd or even
           lda  #0                 ; Index to bit table for even rows
           bcc  sn1                ;
           lda  #2                 ; Index to bit table for odd rows
sn1        sta  tmpcnt             ; Store index to bit table
           ldy  valtabidx+19,x     ; Get index to VIC register values table
           lda  vicvalw,y          ; Get x-coord. (lower 8 bits) for mine sprite
           sta  vicvalw-21,y       ; Set x-coord. for sprite in next row too
           cpx  #10                ; Is this the 2nd mine sprite of this row?
           ldx  tmpcnt             ;
           bcc  sn2                ; No
           inx                     ; Yes, adjust index to bit table
           iny                     ; Also adjust index to 'vicvalw' table
sn2        lda  vicvalw-8,y        ; Current v+16 value for this sprite row
           sec                     ;
           and  oramask1,x         ; Isolate the correct bit
           bne  sn3                ; MSB should be set
           clc                     ; MSB should be cleared
sn3        lda  vicvalw-8-21,y     ; Current v+16 value for the next row
           and  andmask2,x         ; Mask bits according to odd/even rows
           bcc  sn4                ; Do not set MSB for the sprite
           ora  oramask2,x         ;
sn4        sta  vicvalw-8-21,y     ; Store new v+16 value
           lda  vicvalw-8-42,y     ; Current v+16 value for the row below next
           and  andmask2,x         ; Mask bits according to odd/even rows
           bcc  sn5                ; Do not set MSB for the sprite
           ora  oramask2,x         ;
sn5        sta  vicvalw-8-42,y     ; Store new v+16 value
           ldx  minetabidx         ; Restore mine table index
sn6        ldy  #0                 ; Restore sprite image index
           jmp  mm5                ; Draw the mine

; ******************************************************************************
; *  Erase a mine after hitting a sub                                          *
; *  Inputs:  a       Mine sprite number                                       *
; *           x       Mine index                                               *
; ******************************************************************************

erasemine  tay                     ; Transfer mine sprite number to y
           sta  islastr            ; Range is 0-19, so bit 7 will be cleared
           cmp  #10                ; Determine sprite row number
           bcc  em1                ;
           sbc  #10                ;
em1        cmp  #9                 ; Is the mine in the last sprite row?
           bne  em2                ; No
           ror  islastr            ; Yes, set bit 7 to indicate it
em2        lda  mspdatalo,y        ; Get lo byte of sprite image address
           sta  va                 ;
           lda  mspdatahi,y        ; Get hi byte of sprite image address
           sta  va+1               ;
           lda  mineypos,x         ; Get mine's y-coordinate within sprite
           asl  a                  ; Multiply by 2
           adc  mineypos,x         ; Value is now multiplied by 3
           adc  minespoffs,x       ; Add byte offset within sprite image
           tay                     ; Actual sprite image byte index
           lda  minebitidx,x       ; Get index to bit pattern tables
           tax                     ;
           lda  #3-1               ; Height of mine = 3 pixels
           sta  tmpcnt             ;
em3        cpy  #63                ; Need to adjust index further if the mine
           bcc  em4                ; has reached a new sprite row (image)
           cpy  #66                ;
           bcs  em4                ;
           bit  islastr            ; Is this the last sprite row?
           bmi  em5                ; Yes, do not continue erasing this mine
           iny                     ; No, adjust index for next sprite image
em4        lda  (va),y             ;
           and  mineandval,x       ; Erase mine pixels
           sta  (va),y             ;
           iny                     ; Adjust index
           iny                     ;
           iny                     ;
           dec  tmpcnt             ;
           bpl  em3                ; Loop until the entire mine has been erased
em5        ldx  minenum            ; Restore mine index
           rts

; ******************************************************************************
; *  Add points to score for destroying sub                                    *
; *  Inputs:  a       Bit 7 set if a "fast sub" has been destroyed             *
; *           x       Submarine index                                          *
; ******************************************************************************

addpts     asl  a                  ; Is this a "fast sub"?
           bcc  ap4                ; No
           lda  #1*8               ; Yes, 1000 pts for destroying it
           pha                     ;
           lda  #0                 ;
           tax                     ;
           beq  ap5                ; Branch always
ap4        lda  #0                 ; 250 pts for destroying an ordinary sub
           pha                     ;
           lda  #2*8               ;
           ldx  #5*8               ;
ap5        sta  pts+3              ;
           stx  pts+4              ;
           pla                     ;
           sta  pts+2              ;
           clc                     ;
           ldx  #5-1               ; 5 score digits to update
ap1        lda  score,x            ; Get current digit
           bcc  ap2                ;
           adc  #7                 ; Adjust for carry from previous digit
ap2        adc  pts,x              ; Increment current digit
           cmp  #80                ; Overflow?
           bcc  ap3                ; No
           sbc  #80                ; Yes, adjust value
ap3        sta  score,x            ; Store digit
           dex                     ;
           bpl  ap1                ; Loop until done
           rts

; ******************************************************************************
; *  Get one byte of level data and increment address                          *
; ******************************************************************************

getdata    lda  (leveldata),y      ; Get byte
           jmp  $e41b              ; Use KERNAL ROM routine to increment address

; ******************************************************************************
; *  Set up title screen                                                       *
; ******************************************************************************

; Set x-coordinates for title screen sprites in 'vicvalw' table
dsptitle   lda  #lo vicvalw+1+21   ;
           sta  vb                 ;
           lda  #hi vicvalw+1+21   ;
           sta  vb+1               ;
           ldx  #9                 ; 9 sprite rows to process
sl1        txa                     ;
           lsr  a                  ; Determine if row number is odd or even
           ldy  #1                 ;
           lda  #0                 ;
           sta  (vb),y             ; Set v+16 value for this sprite row
           ldy  #4                 ;
           lda  #232               ;
           sta  (vb),y             ; Set x-coordinate for sprite #5
           iny                     ;
           lda  #208               ;
           sta  (vb),y             ; Set x-coordinate for sprite #4
           ldy  #8                 ;
           lda  #136               ;
           bcc  sl2                ;
           lda  #184               ;
sl2        sta  (vb),y             ; Set x-coordinate for sprite #1 or #3
           iny                     ;
           lda  #112               ;
           bcc  sl3                ;
           lda  #160               ;
sl3        sta  (vb),y             ; Set x-coordinate for sprite #0 or #2
           lda  vb                 ;
           clc                     ;
           adc  #21                ; Adjust address
           sta  vb                 ;
           bcc  sl4                ;
           inc  vb+1               ;
sl4        dex                     ;
           bne  sl1                ; Loop until done
; Set up sprite pointers for title screen
           lda  #hi $7bf8          ; Hi byte of sprite pointers address
           ldx  #88-1              ; 88 sprite pointers to set
           ldy  #hi titlespr       ; Hi byte of pointer data address
           jsr  setsprptr          ;
; Set up 'vicvalw' table for title screen
           clc                     ;
           jsr  updvicvalw         ;
; Set up demo data for title screen
           lda  #lo demo           ; Lo byte of address
           sta  leveldata          ;
           lda  #hi demo           ; Hi byte of address
           sta  leveldata+1        ;

; Draw 3 ships into "ships left" sprite images
           ldx  #17-1              ;
dt1        lda  shipgfx,x          ;
           sta  $4000+108*64,x     ;
           sta  $4000+108*64+27,x  ;
           sta  $4000+109*64,x     ;
           dex                     ;
           bpl  dt1                ;

           txa                     ; Acc. is now 255
           ldx  #19-1              ;
ss2        sta  objexpcnt,x        ; No objects exploding/on fire
           dex                     ;
           bpl  ss2                ;

           jsr  resetsubs          ; Reset all submarines etc.

           lda  #%10000000         ;
           sta  gstatus            ; Set status to "Title screen"
           stx  freezedsp          ; Do not freeze display
           rts

; ******************************************************************************
; *  Set up a new game                                                         *
; ******************************************************************************

; Set up sprite pointers for game screen
newgame    lda  #hi $77f8          ; Hi byte of sprite pointers address
           ldx  #80-1              ; 80 sprite pointers to set
           ldy  #hi gamespr        ; Hi byte of pointer data address
           jsr  setsprptr          ;
; Set up 'vicvalw' table for game screen
           sec                     ;
           jsr  updvicvalw         ;
; Set up first level
           lda  #0                 ; Reset level number
           sta  levelnum           ;
           lda  #lo levels         ; Lo byte of address (trigger for level 1)
           sta  leveldata          ;
           lda  #hi levels         ; Hi byte of address (trigger for level 1)
           sta  leveldata+1        ;

           ldx  #2                 ;
           stx  ships              ; Number of ships left
           jsr  setupship          ; Set up ship's x-coord. and sprite images
           jsr  resetsubs          ; Reset all submarines etc.

resetgame  ldx  #6-1               ;
           lda  #0*8+7             ;
rg1        sta  score,x            ; Reset all score digits
           dex                     ;
           bpl  rg1                ;
           inx                     ; x is now zero
           stx  mineupdcnt         ; Reset counter for update of mines on screen
           rts

; ******************************************************************************
; *  Reset all submarines etc.                                                 *
; ******************************************************************************

resetsubs  ldx  #18-1              ;
           stx  subrelint          ; Reset sub release interval counter
re1        lda  #128               ;
           sta  objxposmsb,x       ; Disable all submarines
           dex                     ;
           bpl  re1                ;
           stx  fastsub            ; No "fast sub" initially
           stx  dropdelay          ; Initial drop delay
           inx                     ; x is now zero
           stx  endoflevel         ; Reset end-of-level indicator
           stx  subrelcnt          ; Reset sub release counter
           stx  torpedocnt         ; No torpedo moving initially
           stx  isshiplost         ; No ship lost initially
           rts

; ******************************************************************************
; *  Set up ship's x-coordinates and sprite images, reset number of mines etc. *
; *  Inputs:  x       Number of ships left (0-2)                               *
; ******************************************************************************

setupship  ldy  shipspridx,x       ; Get index
           ldx  #17-1              ;
           lda  #0                 ;
ss1        sta  $4000+108*64,y     ; Erase one ship from "ships left" sprites
           iny                     ;
           dex                     ;
           bpl  ss1                ;

           sta  objxposmsb+18      ; Initial MSB of virtual x-coordinate
           lda  #224               ;
           sta  objxpos+18         ; Initial virtual x-coordinate of ship
           lda  #160               ;
           sta  vicvalw+216        ; Actual VIC x-coordinate for left sprite
           lda  #184               ;
           sta  vicvalw+215        ; Actual VIC x-coordinate for right sprite
           lda  vicvalw+212        ; Current v+16 value for the top sprite row
           and  #%11001111         ; Turn off MSB for ship sprites
           sta  vicvalw+212        ; Store new v+16 value
           ldx  #30                ;
           stx  $53f8+4            ; Set sprite pointer for left part of ship
           inx                     ;
           stx  $53f8+5            ; Set sprite pointer for right part of ship

; Reset number of mines to 32
resetmines lda  #3*8+7             ; Byte index to 'charset' table
           sta  mines+1            ;
           lda  #2*8+7             ; Byte index to 'charset' table
           sta  mines+2            ;
           rts

; ******************************************************************************
; *  Increment level number and prepare for displaying it                      *
; ******************************************************************************

; x is assumed to be 255 here
inclevel   stx  level+2            ; Default: 2nd digit is a blank char
           ldx  #0                 ; Reset index
           sed                     ; Enable decimal mode
           lda  levelnum           ; Get current level number (BCD format)
           clc                     ;
           adc  #1                 ; Increment it
           cmp  #17                ; Has it reached 11? (BCD format)
           bne  il1                ; No
           lda  #lo levels+1       ; Lo byte of address (level 1)
           sta  leveldata          ;
           lda  #hi levels+1       ; Hi byte of address (level 1)
           sta  leveldata+1        ;
           lda  #1                 ; Start from level 1 again
il1        sta  levelnum           ; Store new value
           cld                     ; Disable decimal mode
           and  #%11110000         ; Isolate 1st BCD digit
           beq  il2                ; It's zero - do not display it on screen
           lsr  a                  ; 1st digit is now multiplied by 8
           adc  #7                 ; Calculate final byte index to 'charset'
           sta  level+1,x          ; Store it
           inx                     ; Adjust index
il2        lda  levelnum           ; Get level number
           and  #%00001111         ; Isolate 2nd BCD digit
           asl  a                  ; Multiply by 8
           asl  a                  ;
           asl  a                  ;
           adc  #7                 ; Calculate final byte index to 'charset'
           sta  level+1,x          ; Store it
           rts

; ******************************************************************************
; *                         I n i t i a l   s e t u p                          *
; ******************************************************************************

setup      sei                     ; Disable interrupts
           lda  #$7f               ;
           sta  $dc0d              ; Disable timer interrupts
           lda  $dc0d              ; Acknowledge CIA #1 interrupts
           jsr  setirqa            ; Set irq vector for the first interrupt
           lda  #$86               ;
           sta  nmivector          ; Set lo byte of NMI routine
           lda  #$ea               ;
           sta  nmivector+1        ; Set hi byte of NMI routine
           lda  #$81               ;
           sta  $dd0d              ; Enable CIA #2 Timer A interrupt
           ldx  #0                 ;
           stx  $dd05              ; Prepare timer to count from 1 to 0
           inx                     ;
           stx  $dd04              ;
           lda  #$dd               ; Cause an NMI interrupt for
           sta  $dd0e              ; restore key protection

; ******************************************************************************
; *  Clear memory locations $4000 - $7fff inclusive                            *
; ******************************************************************************

           lda  #0                 ; Lo byte and value to store in memory
           sta  va                 ;
           ldx  #64                ; Hi byte and number of pages
           stx  va+1               ;
           tay                     ;
se1        sta  (va),y             ;
           iny                     ;
           bne  se1                ;
           inc  va+1               ; Next page
           dex                     ;
           bne  se1                ; Loop until done

; ******************************************************************************
; *  Relocate data                                                             *
; ******************************************************************************

; y is assumed to be zero here
           lda  #lo relocate       ; Start address of data to be relocated
           sta  va                 ;
           lda  #hi relocate       ;
           sta  va+1               ;
de1        lda  (va),y             ; Get a byte
           cmp  #endofdata         ; Code for end of data?
           beq  fl1                ; Yes
           jsr  $fcdb              ; Use KERNAL ROM routine to inc 'va'
           cmp  #new_addr          ; Code for new destination address?
           bne  de2                ; No
           lda  (va),y             ; Yes, get lo byte of new address
           sta  leveldata          ; Store it
           jsr  $fcdb              ; Use KERNAL ROM routine to inc 'va'
           lda  (va),y             ; Get hi byte of new address
           sta  leveldata+1        ; Store it
           jsr  $fcdb              ; Use KERNAL ROM routine to inc 'va'
           bne  de1                ; Branch always
de2        sta  (leveldata),y      ; Store data byte
           jsr  $e41b              ; Use KERNAL ROM routine to inc 'leveldata'
           bne  de1                ; Branch always

; ******************************************************************************
; *  Create sprite images                                                      *
; ******************************************************************************

; Flip graphics data horizontally
fl1        ldx  #240               ; 16 sprites * 45 bytes / 3 = 240
fl2        lda  tmpspdata-1,x      ;
           jsr  flipbyte           ;
           sta  tmpspdata+719,x    ;
           lda  tmpspdata+239,x    ;
           jsr  flipbyte           ;
           sta  tmpspdata+959,x    ;
           lda  tmpspdata+479,x    ;
           jsr  flipbyte           ;
           sta  tmpspdata+1199,x   ;
           dex                     ;
           bne  fl2                ;
           
           jsr  gfxtospr           ; Copy graphics data to actual sprite images

; Create sprite for water surface (to hide the fact that the background
; color register is actually changed a little too late)
           lda  #%10101010         ;
           sta  $4000+76*64+39     ;
           sta  $4000+76*64+40     ;
           sta  $4000+76*64+41     ;

           jsr  setupsptxt         ; Write various game messages into sprites
           jsr  setuplogo          ; Write game logo into sprites

; Create 'vicidxr' table
           sec                     ;
           ldx  #210               ;
sv4        lda  #30                ;
           sta  vicidxr-19,x       ;
           txa                     ;
           sbc  #21                ;
           tax                     ;
           bne  sv4                ;

; Initialize mine sprite tables
           ldx  #20-1              ;
           lda  #0                 ;
ss3        sta  mspxpos,x          ;
           sta  mspxposmsb,x       ;
           sta  mspminecnt,x       ; No mines in any of the mine sprites
           dex                     ;
           bpl  ss3                ;
           sta  txttabidx          ; Reset index to "text in sprites" tables

; Initialize various SID and VIC registers
           ldx  #24                ;
ss4        sta  s,x                ; Reset SID registers
           dex                     ;
           bpl  ss4                ;
           lda  #%00000001         ;
           sta  s+23               ; Filter voice 1 output (sub explosions)
           lda  #%10011111         ; Max volume, low-pass filter and
           sta  s+24               ; disable voice 3 output
           lda  #10                ;
           sta  s+5                ; Set attack/decay for voice 1
           lda  #12                ;
           sta  s+12               ; Set attack/decay for voice 2
           sta  v+37               ; Set sprite multicolor #0 (grey)
           lda  #15                ;
           sta  s+22               ; Set filter cutoff frequency hi byte
           sta  v+38               ; Set sprite multicolor #1 (light grey)
           lsr  a                  ;
           sta  v+45               ; Set color of sprite #6 (yellow)
           lda  #128               ;
           sta  s+18               ; Set noise waveform for voice 3
           stx  s+14               ; Set lo byte of voice 3 frequency
           stx  s+15               ; Set hi byte of voice 3 frequency
           ldx  #2                 ;
           stx  s+1                ; Set hi byte of voice 1 frequency
           stx  s+8                ; Set hi byte of voice 2 frequency

; Acc. is assumed to be 128 here
           ldx  #10-1              ;
ss5        sta  minespnum,x        ; All mines initially disabled
           dex                     ;
           bpl  ss5                ;
           stx  v+21               ; Enable all sprites

           jsr  resetgame          ; Reset score etc.
           jsr  resetmines         ; Reset number of mines

           ldy  #2                 ;
           sty  $dd00              ; Set VIC bank 1
           lda  #21                ;
           sta  v+18               ; Set raster line for the first interrupt
           lda  #249               ;
ss6        cmp  v+18               ;
           bne  ss6                ; Wait for raster line 249
           dey                     ; y is now 1
           sty  v+17               ; Disable lower border
           sty  v+25               ; Acknowledge interrupt
           sty  v+26               ; Set raster interrupts to occur
           cli                     ; Enable interrupts
           rts

; ******************************************************************************
; *  Flip a multicolor graphics byte horizontally                              *
; ******************************************************************************

flipbyte   ldy  #4-1               ; Four bit pairs to swap
sb1        lsr  a                  ; Get first bit of this bit pair
           php                     ; Push on stack
           lsr  a                  ; Get next bit
           rol  tmpa               ; Rotate into destination byte
           plp                     ; Pull first bit from stack
           rol  tmpa               ; Rotate into destination byte
           dey                     ;
           bpl  sb1                ; Loop until done
           lda  tmpa               ;
           rts

; ******************************************************************************
; *  Copy graphics data to the actual sprite images                            *
; ******************************************************************************

; x is assumed to be zero here
gfxtospr   stx  va                 ; Set lo byte of source address
           lda  #hi tmpspdata      ;
           sta  va+1               ; Set hi byte of source address
           lda  #lo $4000+30*64    ;
           sta  leveldata          ; Set lo byte of destination address
           lda  #hi $4000+30*64    ;
           sta  leveldata+1        ; Set hi byte of destination address

           ldy  #30                ; Normal, not flipped graphics data
           jsr  copygfx            ; Copy graphics data

           ldx  #30                ;
           ldy  #0                 ; Flipped graphics data

; ******************************************************************************
; *  Copy graphics data to sprites                                             *
; *  Inputs:  x       0 for normal, 30 for flipped data                        *
; *           y       30 for normal, 0 for flipped data                        *
; ******************************************************************************

copygfx    lda  #16-1              ; 16 sprites to process
           sta  tmpa               ; Number of sprites to process
           stx  tmpc               ; Index to 1st byte to copy
           sty  tmpd               ; Index to 3rd byte to copy
           ldx  #0                 ;
cg1        lda  #15-1              ; 15 lines per sprite
           sta  tmpb               ;
cg2        ldy  tmpc               ; Index to 1st byte
           jsr  copybyte           ; Copy graphics byte to sprite image
           ldy  #15                ; Index to 2nd byte
           jsr  copybyte           ; Copy graphics byte to sprite image
           ldy  tmpd               ; Index to 3rd byte
           jsr  copybyte           ; Copy graphics byte to sprite image
           jsr  $fcdb              ; Use KERNAL ROM routine to inc 'va'
           dec  tmpb               ;
           bpl  cg2                ; Loop until current sprite is done
           lda  #30                ;
           jsr  addtova            ; Adjust graphics data address
           lda  leveldata          ;
           clc                     ;
           adc  #19                ; Adjust sprite image address
           sta  leveldata          ;
           bcc  cg3                ;
           inc  leveldata+1        ;
cg3        dec  tmpa               ;
           bpl  cg1                ; Loop until all sprites have been done
           rts

; ******************************************************************************
; *  Copy graphics byte to sprite image                                        *
; ******************************************************************************

copybyte   lda  (va),y             ; Get graphics byte
           sta  (leveldata,x)      ; Store in sprite image
           jmp  $e41b              ; Use KERNAL ROM routine to inc 'leveldata'

; ******************************************************************************
; *  Set sprite pointers for several rows                                      *
; *  Inputs:  a       Hi byte of sprite pointers address                       *
; *           x       Number of sprite pointers to set - 1                     *
; *           y       Hi byte of pointer data address                          *
; ******************************************************************************

setsprptr  sta  vb+1               ;
           sty  sp2+2              ; Self-modifying code
           lda  #$f8               ; Lo byte of sprite pointers address
           sta  vb                 ;
sp1        ldy  #8-1               ;
sp2        lda  titlespr,x         ; Set 8 sprite pointers per row
           sta  (vb),y             ;
           dex                     ;
           dey                     ;
           bpl  sp2                ;
           lda  vb+1               ;
           sec                     ;
           sbc  #4                 ; Adjust hi byte of address
           sta  vb+1               ;
           txa                     ;
           bpl  sp1                ; Loop until done
           rts

; ******************************************************************************
; *  Write game messages and scoring area text into sprite images              *
; ******************************************************************************

; "PRESS FIRE TO PLAY" and "   (C) 2004"
setupsptxt lda  #39                ; End index to text
           sta  tmpa               ;
           lda  #lo $4000+80*64    ; Lo byte of address
           ldx  #hi $4000+80*64    ; Hi byte of address
           ldy  #12                ; Start index to text
           jsr  writesptxt         ;
; "GEIR STRAUME"
           lda  #51                ; End index to text
           sta  tmpa               ;
           lda  #lo $4000+86*64+33 ; Lo byte of address
           ldx  #hi $4000+86*64+33 ; Hi byte of address
           jsr  writesptxt         ;
; "SCORE  HIGH MINES ", "GAME OVER" and " LEVEL"
           lda  #84                ; End index to text
           sta  tmpa               ;
           lda  #lo $4000+96*64    ; Lo byte of address
           ldx  #hi $4000+96*64    ; Hi byte of address

; ******************************************************************************
; *  Write text into sprite images                                             *
; *  Inputs:  a       Lo byte of start sprite image address                    *
; *           x       Hi byte of start sprite image address                    *
; *           y       Start index to text table                                *
; *           tmpa    End index to text table                                  *
; ******************************************************************************

writesptxt sta  va                 ; Store lo byte of sprite image address
           stx  va+1               ; Store hi byte of sprite image address
pr1        lda  #3-1               ; 3 characters per sprite
           sta  tmpb               ;
pr2        sty  tmpc               ; Start index
           ldx  sptxtdata,y        ; Get character
           beq  pr4                ; This is a space character, so skip it
           ldy  #21                ;
pr3        lda  charset,x          ; Transfer character data to sprite image
           sta  (va),y             ;
           dex                     ;
           dey                     ;
           dey                     ;
           dey                     ;
           bpl  pr3                ; Loop until done
           ldy  tmpc               ;
pr4        iny                     ; Next character
           inc  va                 ;
           dec  tmpb               ;
           bpl  pr2                ; Loop for all chars in this sprite
           lda  #61                ;
           jsr  addtova            ; Adjust sprite image address
           cpy  tmpa               ; End index reached?
           bne  pr1                ; No, loop
           rts

; ******************************************************************************
; *  Write game logo into sprite images                                        *
; ******************************************************************************

; "SUB"
setuplogo  lda  #3                 ; End index to text
           sta  tmpa               ;
           lda  #lo $4000+112*64+42 ; Lo byte of address
           ldx  #hi $4000+112*64+42 ; Hi byte of address
           ldy  #0                 ; Start index to text
           jsr  writelogo          ;
; "DESTROYER"
           lda  #12                ; End index to text
           sta  tmpa               ;
           lda  #lo $4000+114*64+42 ; Lo byte of address
           ldx  #hi $4000+114*64+42 ; Hi byte of address

; ******************************************************************************
; *  Write logo text (big characters) into sprite images                       *
; *  Inputs:  a       Lo byte of start sprite image address                    *
; *           x       Hi byte of start sprite image address                    *
; *           y       Start index to text table                                *
; *           tmpa    End index to text table                                  *
; ******************************************************************************

writelogo  sta  va                 ; Lo byte of start sprite image address
           stx  va+1               ; Hi byte of start sprite image address
           ldx  #2                 ;
lo1        stx  vb                 ; Store index to 'rightpart' and 'nextchar'
           sty  tmpb               ; Store index to game title characters
           ldx  sptxtdata,y        ; Get last byte index for game title char
           lda  #8-1               ; Process all bytes in the character
           sta  vb+1               ;
lo2        stx  tmpcnt             ; Store byte index to char
           lda  charset,x          ; Get a byte
           ldx  #8-1               ; Process all bits in the byte
lo3        asl  a                  ; Get a bit
           php                     ; Push on stack
           rol  tmpd               ; Rotate into the right character byte
           rol  tmpc               ; Rotate the left character byte
           plp                     ; Pull bit from stack
           rol  tmpd               ; Rotate into the right character byte
           rol  tmpc               ; Rotate the left character byte
           dex                     ;
           bpl  lo3                ; Loop until all 8 bits have been processed
           lda  tmpc               ; Get left part of character byte
           ldy  #0                 ;
           sta  (va),y             ; Store in sprite image
           ldy  #3                 ;
           sta  (va),y             ; Store again
           lda  tmpd               ; Get right part of character byte
           ldx  vb                 ; Get index to 'rightpart' and 'nextchar'
           ldy  rightpart,x        ;
           sta  (va),y             ; Store in sprite image
           iny                     ;
           iny                     ;
           iny                     ;
           sta  (va),y             ; Store again
           lda  va                 ;
           sec                     ;
           sbc  #6                 ; Adjust address
           sta  va                 ;
           bcs  lo4                ;
           dec  va+1               ;
lo4        ldx  tmpcnt             ; Get byte index to char
           dex                     ; Decrement it
           dec  vb+1               ;
           bpl  lo2                ; Loop until all 8 bytes have been processed
           ldx  vb                 ; Get index to 'rightpart' and 'nextchar'
           lda  nextchar,x         ;
           jsr  addtova            ; Adjust sprite image address
           dex                     ; Decrement index to tables
           bpl  lo6                ;
           ldx  #2                 ; Reset table index every third char
lo6        ldy  tmpb               ; Get index to game title characters
           iny                     ; Increment it
           cpy  tmpa               ; End index reached?
           bne  lo1                ; No, loop
           rts

; ******************************************************************************
; *  Set up three sprites for displaying in-game message                       *
; *  Inputs:  x       First sprite image                                       *
; ******************************************************************************

setmsgspr  stx  $67f8              ; Set sprite pointer
           inx                     ;
           stx  $67f9              ; Set sprite pointer
           inx                     ;
           stx  $67fa              ; Set sprite pointer
           lda  #148               ;
           sta  vicvalw+136        ; Set x-coordinates
           lda  #172               ;
           sta  vicvalw+135        ;
           lda  #196               ;
           sta  vicvalw+115        ;
           lda  vicvalw+107        ; Current v+16 value for the sprite row
           and  #%11110000         ; Turn off MSB for message sprites
           sta  vicvalw+107        ; Store new v+16 value
           clc                     ;
           ldy  #3                 ; To set white color for message sprites
           db   $2c                ; Skip next two bytes using a bit instruction

; ******************************************************************************
; *  Update 'vicvalw' table                                                    *
; *  Inputs:  C       Set for game screen, not set for title screen            *
; ******************************************************************************

updvicvalw ldy  #13                ;
uv1        ldx  vicvalwidx-1,y     ; Get index
           lda  vicvalwval-1,y     ; Get value for game screen
           bcs  uv2                ;
           lda  vicvalwval+12,y    ; Get value for title screen
uv2        sta  vicvalw,x          ; Store in 'vicvalw' table
           dey                     ;
           bne  uv1                ; Loop until done
           rts

; ******************************************************************************
; *  Add byte value to 'va' variable                                           *
; *  Inputs:  a       Value to add                                             *
; ******************************************************************************

addtova    clc                     ;
           adc  va                 ; Add to lo byte
           sta  va                 ;
           bcc  av1                ;
           inc  va+1               ; Adjust hi byte
av1        rts

; ******************************************************************************
; *  Set irq vector for the first interrupt routine                            *
; ******************************************************************************

setirqa    lda  #lo irqa           ;
           sta  irqvector          ; Set lo byte of first interrupt routine
           lda  #hi irqa           ;
           sta  irqvector+1        ; Set hi byte of first interrupt routine
           rts

; ******************************************************************************
; *                          V a r i o u s   d a t a                           *
; ******************************************************************************

; ################## Data for writing game logo into sprites ###################
; Sprite image index to the right part of current character
rightpart  db   1,62,1
; Sprite image index to the left part of next character
nextchar   db   111,111,50

; ############################### Display data #################################
; Indexes to the 'vicvalw' table
; - X-coordinates of the RIGHT sub sprites (0-17)
valtabidx  db   194,192,173,171,152,150,131,129,110,108,89,87,68,66,47,45,26,24
; - X-coordinate of the RIGHT ship sprite (18)
           db   215
; - X-coordinates of mine sprites (19-38)
           db   220,199,178,157,136,115,94,73,52,31 ; 1st mine sprite of a row
           db   219,198,177,156,135,114,93,72,51,30 ; 2nd mine sprite of a row

; Sprite images for left and right part of ship (normal/on fire)
shipspimgl db   30,47,32,49,34,51
shipspimgr db   31,46,33,48,35,50

; Index for erasing a ship in "ships left" sprite images
shipspridx db   0,64,27

; Bit pattern for "ships left" sprite images
shipgfx    db   %00000000,%01110000,%00000000
           db   %00000000,%01110000,%00000000
           db   %00000100,%01110001,%01000000
           db   %01010101,%01111111,%11110000
           db   %11111111,%11111111,%11000000
           db   %00111111,%11111111

; ******************************************************************************
; *                            L e v e l   d a t a                             *
; ******************************************************************************

; First data byte: bit 7 set     = new level
;                  bit 7 not set = new sub sequence (see below)
;
; Sub sequence data:
;
; %01+(count*4),interval: One sub per row (sprites #4 and #5), random sub index
;                         (torpedo launch mode)
;
; %11+(count*4),interval: One sub per row (sprites #6 and #7), random sub index
;                         (title screen demo mode)
;
; %10+(count*4),interval: Two subs per row, random sub index
;
; %00+(count*4):          Release of a specific, fast sub (fixed mode)
;                         Two bytes per sub:
;                         (dirleft*1)+(torpedo*2)+(sub*4),interval
;
; Notes:
;
; 1. Torpedoes use sprite #6, so no subs using sprites #6 and #7 (odd sub
;    numbers) should be on-screen when a torpedo is launched
; 2. Value of 'count' is 5 bits (range 0-31)
; 3. If 'interval' is zero, no subs are released ('interval' is set to 256)
;    (this does not apply to fixed mode)
; 4. Minimum 'interval' between two subs in same row (mode %10) should be 121
;    to avoid "overlapping" subs
; 5. Valid range of 'sub' is 0-17

; ######################### Demo data for title screen #########################
demo       db   %11+(31*4),121
levels     db   %11111111   ; Restart demo (also triggers the first level)
; ################################## Level 1 ###################################
           db   %01+(24*4),121
           db   %01+(4*4),0 ; Wait for subs to exit screen
           db   %11111111   ; Next level
; ################################## Level 2 ###################################
           db   %00+(1*4),(1*1)+(0*2)+(17*4),121
           db   %10+(10*4),121
           db   %11+(4*4),0 ; Wait for subs to exit screen
           db   %01+(12*4),91
           db   %01+(4*4),0 ; Wait for subs to exit screen
           db   %11111111   ; Next level
; ################################## Level 3 ###################################
           db   %00+(1*4),(0*1)+(1*2)+(8*4),61
           db   %01+(20*4),61
           db   %01+(4*4),0 ; Wait for subs to exit screen
           db   %11111111   ; Next level
; ################################## Level 4 ###################################
           db   %01+(10*4),91
           db   %00+(1*4),(0*1)+(0*2)+(0*4),255
           db   %11+(4*4),0 ; Wait for subs to exit screen
           db   %11111111   ; Next level
; ################################## Level 5 ###################################
           db   %00+(1*4),(1*1)+(1*2)+(4*4),61
           db   %01+(24*4),61
           db   %11+(4*4),0 ; Wait for subs to exit screen
           db   %11111111   ; Next level
; ################################## Level 6 ###################################
           db   %00+(1*4),(1*1)+(0*2)+(1*4),121
           db   %10+(10*4),121
           db   %11+(4*4),0 ; Wait for subs to exit screen
           db   %00+(1*4),(0*1)+(1*2)+(6*4),121
           db   %01+(12*4),91
           db   %01+(4*4),0 ; Wait for subs to exit screen
           db   %11111111   ; Next level
; ################################## Level 7 ###################################
           db   %01+(12*4),91
           db   %01+(16*4),41
           db   %01+(4*4),0 ; Wait for subs to exit screen
           db   %11111111   ; Next level
; ################################## Level 8 ###################################
           db   %00+(1*4),(1*1)+(1*2)+(4*4),61
           db   %01+(24*4),61
           db   %01+(4*4),0 ; Wait for subs to exit screen
           db   %11111111   ; Next level
; ################################## Level 9 ###################################
           db   %00+(1*4),(1*1)+(1*2)+(2*4),61
           db   %01+(24*4),61
           db   %01+(4*4),0 ; Wait for subs to exit screen
           db   %11111111   ; Next level
; ################################## Level 10 ##################################
           db   %00+(1*4),(0*1)+(1*2)+(4*4),255
           db   %01+(1*4),0 ; Wait for subs to exit screen
           db   %00+(1*4),(1*1)+(1*2)+(2*4),255
           db   %01+(30*4),41
           db   %01+(4*4),0 ; Wait for subs to exit screen
           db   %11111111   ; Restart at level 1

; ******************************************************************************
; *                  D a t a   t o   b e   r e l o c a t e d                   *
; ******************************************************************************

; The following special code sequences are used within this data:
;
;      new_addr,lo,hi   Specifies new destination address
;      endofdata        Marks end of data

relocate   db   new_addr,lo zpdata,hi zpdata
; ################################ Bit tables ##################################
; Bit tables for masking v+16 values for mine sprites (0, 1, 2 and 3)
; 'andmask2'
           db   251,247 ; + the first two bytes in 'andmask1'
; 'andmask1'
           db   254,253,251,247
; Bit table for masking v+16 values for sprites 4/5 and 6/7 ('andmask3')
           db   207,0,63,0 ; Only byte 0 and 2 are actually used
; Bit table for setting v+16 values for mine sprites ('oramask2')
           db   4,8 ; + the first two bytes in 'oramask1'
; Bit table for general use ('oramask1')
           db   1,2,4,8,16,32,64,128

; ############## Bit patterns for drawing and erasing mine pixels ##############
; Bit patterns for mines within a sprite (index to table is 'diff and #7', where
; diff is 'mine sprite x-coord.' - 'left ship sprite x-coord.')
; 'mineandval'
           db   252,252,243,243,207,207,63,63
; 'moneoraval'
           db   2+1,2+1,8+4,8+4,32+16,32+16,128+64,128+64

; ############## Tables for writing score etc. into sprite images ##############
; Start indexes to table 'sptxtdata' ('txtstart')
           db   96,84,87,90,93,104
; End indexes to table 'sptxtdata' ('txtend')
           db   99,87,90,93,96,107
; Lo byte of text sprite image address ('txtsplo')
           db   lo $4000+100*64+27,lo $4000+98*64+27,lo $4000+99*64+27
           db   lo $4000+96*64+27,lo $4000+97*64+27,lo $4000+107*64
; Hi byte of text sprite image address ('txtsphi')
           db   hi $4000+100*64+27,hi $4000+98*64+27,hi $4000+99*64+27
           db   hi $4000+96*64+27,hi $4000+97*64+27,hi $4000+107*64

; ################### Text data (byte indexes to 'charset') ####################
           db   new_addr,lo sptxtdata,hi sptxtdata
; "SUBDESTROYER" (game logo in sprites 112-119)
           db   207,223,95,111,119,207,215,199,183,239,119,199
; "PRESS FIRE TO PLAY" (sprites 80-85)
           db   191,199,119,207,207,0,127,151,199,119,0,215,183,0,191,159,87,239
; "   (C) 2004" (sprites 86-88)
           db   0,0,0,247,0,23,7,7,39
; "GEIR STRAUME" (sprites 86-89)
           db   135,119,151,199,0,207,215,199,87,223,167,119
; "SCORE  HIGH MINES " (sprites 96-101)
           db   207,103,183,199,119,0,0,143,151,135,143,0,167,151,175,119,207,0
; "GAME OVER" (sprites 102-104)
           db   135,87,167,119,0,183,231,119,199
; " LEVEL" (sprites 105-107)
           db   0,159,119,231,119,159
; 'hiscore' (index 84-89)
           db   7,15,7,7,7,7
; 'score' (index 90-95)
; 'mines' (index 96-98)
; 'pts'   (index 99-103)
; 'level' (index 104-106)

; ############################ Sprite pointer data #############################
; Sprite pointer data for title screen
           db   new_addr,lo titlespr,hi titlespr
           db   44,44,44,44,44,44,44,76       ; Blank sprites/water surface
           db   44,44,44,44,44,44,44,44       ; Blank sprites
           db   44,44,112,113,44,44,44,44     ; Game logo top
           db   114,115,116,117,118,119,44,44 ; Game logo bottom
           db   44,44,44,44,44,44,44,44       ; Blank sprites
           db   44,44,44,44,44,44,44,44       ; Blank sprites
           db   44,44,44,44,44,44,44,44       ; Blank sprites
           db   44,44,44,44,44,44,44,44       ; Blank sprites
           db   44,86,87,88,89,44,44,44       ; Copyright text
           db   44,44,44,44,44,44,44,44       ; Blank sprites
           db   96,97,98,99,100,101,108,109   ; Score, high, mines, ships left
; Sprite pointer data for game screen
           db   new_addr,lo gamespr,hi gamespr
           db   0,10,44,44,44,44,44,76        ; Mines/water surface
           db   44,44,1,11,44,44,44,44        ; Mines/submarines 1
           db   2,12,44,44,44,44,44,44        ; Mines/submarines 2
           db   44,44,3,13,44,44,44,44        ; Mines/submarines 3
           db   4,14,44,44,44,44,44,44        ; Mines/submarines 4
           db   44,44,5,15,44,44,44,44        ; Mines/submarines 5
           db   6,16,44,44,44,44,44,44        ; Mines/submarines 6
           db   44,44,7,17,44,44,44,44        ; Mines/submarines 7
           db   8,18,44,44,44,44,44,44        ; Mines/submarines 8
           db   44,44,9,19,44,44,44,44        ; Mines/submarines 9

; ###################### Addresses of mine sprite images #######################
; 'mspdatalo'
           db   $00,$40,$80,$c0,$00,$40,$80,$c0,$00,$40
           db   $80,$c0,$00,$40,$80,$c0,$00,$40,$80,$c0
; 'mspdatahi'
           db   $40,$40,$40,$40,$41,$41,$41,$41,$42,$42
           db   $42,$42,$43,$43,$43,$43,$44,$44,$44,$44

; ############# Data for updating sprite colors in 'vicvalw' table #############
; 'vicvalwidx'
           db   126,125,124,113,112,105,210,209,189,188,176,175,229
; 'vicvalwval'
           db   0,0,0,0,7,7,0,0,0,0,7,7,240 ; Game screen
           db   1,1,1,1,1,1,7,7,7,7,7,7,192 ; Title screen

; ################# Data for displaying the main playing area ##################
; VIC register indexes for write ('vicidxw')
           db   new_addr,lo vicidxw+1,hi vicidxw+1
           db   24,16,14,12,10,8,27,32,2,0,15,13,11,9,7,5,3,1,27,27,27
           db   24,16,14,12,10,8,27,27,6,4,15,13,11,9,7,5,3,1,27,27,27
           db   24,16,14,12,10,8,27,27,2,0,15,13,11,9,7,5,3,1,27,27,27
           db   24,16,14,12,10,8,27,27,6,4,15,13,11,9,7,5,3,1,27,27,27
           db   24,16,14,12,10,8,27,27,2,0,15,13,11,9,7,5,3,1,27,27,44
           db   24,16,14,12,10,8,43,42,6,4,15,13,11,9,7,5,3,1,41,40,39
           db   24,16,14,12,10,8,27,27,2,0,15,13,11,9,7,5,3,1,27,27,27
           db   24,16,14,12,10,8,27,27,6,4,15,13,11,9,7,5,3,1,27,27,27
           db   24,16,14,12,10,8,44,43,2,0,15,13,11,9,7,5,3,1,27,42,41
           db   24,16,14,12,10,8,46,33,6,4,15,13,11,9,7,5,3,1,46,40,39
; The top sprite row ('vicidxwtop')
           db   24,16,14,12,10,8,6,4,2,0,15,13,11,9,7,5,3,1,28
; The bottom sprite row ('vicidxwbtm')
           db   28,6,4,2,0,16,17

; VIC register values to be written ('vicvalw')
           db   new_addr,lo vicvalw+1,hi vicvalw+1
           db   224,0,64,40,240,216,0,6,0,0,0,0,0,0,0,0,0,0,0,0,0
           db   208,0,0,0,0,0,0,0,0,0,213,213,213,213,213,213,213,213,0,0,0
           db   192,0,0,0,0,0,0,0,0,0,192,192,192,192,192,192,192,192,0,0,0
           db   176,0,0,0,0,0,0,0,0,0,171,171,171,171,171,171,171,171,0,0,0
           db   160,0,0,0,0,0,0,0,0,0,150,150,150,150,150,150,150,150,0,0,0
           db   144,0,0,0,0,0,0,0,0,0,129,129,129,129,129,129,129,129,0,0,0
           db   128,0,0,0,0,0,0,0,0,0,108,108,108,108,108,108,108,108,0,0,0
           db   112,0,0,0,0,0,0,0,0,0,87,87,87,87,87,87,87,87,0,0,0
           db    96,0,0,0,0,0,0,0,0,0,66,66,66,66,66,66,66,66,0,0,0
           db    80,0,0,0,0,0,7,6,0,0,45,45,45,45,45,45,45,45,6,6,6
; The top sprite row ('vicvalwtop')
           db    64,128,236,0,0,0,0,0,0,0,24,24,24,24,24,24,24,24,192
; The bottom sprite row ('vicvalwbtm')
           db   192,144,120,48,24,192,0

; ################## Bit patterns for alphanumeric characters ##################
           db   new_addr,lo charset,hi charset
; "0" (7) (index to last byte of char)
           db   %00101000
           db   %01001100
           db   %11000110
           db   %11000110
           db   %11000110
           db   %01101100
           db   %00101000
           db   %00000000
; "1" (15)
           db   %00011000
           db   %00111000
           db   %00011000
           db   %00011000
           db   %00011000
           db   %00011000
           db   %01111110
           db   %00000000
; "2" (23)
           db   %01011100
           db   %11000110
           db   %00001110
           db   %00111100
           db   %01110000
           db   %11100110
           db   %11101110
           db   %00000000
; "3" (31)
           db   %11011110
           db   %10001100
           db   %00011000
           db   %00111100
           db   %00000110
           db   %11000110
           db   %01011100
           db   %00000000
; "4" (39)
           db   %00001100
           db   %00101100
           db   %01101100
           db   %11001100
           db   %11101110
           db   %00001100
           db   %00011110
           db   %00000000
; "5" (47)
           db   %11111100
           db   %11000000
           db   %11111100
           db   %00000110
           db   %00000110
           db   %11000110
           db   %01011100
           db   %00000000
; "6" (55)
           db   %00110110
           db   %01100010
           db   %11000000
           db   %11111100
           db   %11000110
           db   %11000110
           db   %01110100
           db   %00000000
; "7" (63)
           db   %11101110
           db   %11000110
           db   %00001100
           db   %00011000
           db   %00110000
           db   %00110000
           db   %00110000
           db   %00000000
; "8" (71)
           db   %01101000
           db   %11000100
           db   %11100100
           db   %01111000
           db   %10011110
           db   %10000110
           db   %01101100
           db   %00000000
; "9" (79)
           db   %01011100
           db   %11000110
           db   %11000110
           db   %01110110
           db   %00000110
           db   %00001100
           db   %01111000
           db   %00000000
; "A" (87)
           db   %00001100
           db   %00011100
           db   %00101100
           db   %00100110
           db   %01011110
           db   %01000110
           db   %11100110
           db   %00000000
; "B" (95)
           db   %11101100
           db   %01100110
           db   %01100110
           db   %01101100
           db   %01100110
           db   %01100110
           db   %11101100
           db   %00000000
; "C" (103)
           db   %00101100
           db   %01100110
           db   %11000000
           db   %11000000
           db   %11000000
           db   %01100110
           db   %00101100
           db   %00000000
; "D" (111)
           db   %11101100
           db   %01101110
           db   %01100110
           db   %01100110
           db   %01100110
           db   %01101110
           db   %11101100
           db   %00000000
; "E" (119)
           db   %11101110
           db   %01100110
           db   %01100000
           db   %01101100
           db   %01100000
           db   %01100110
           db   %11101110
           db   %00000000
; "F" (127)
           db   %11101110
           db   %01100110
           db   %01100000
           db   %01101100
           db   %01100000
           db   %01100000
           db   %11100000
           db   %00000000
; "G" (135)
           db   %00110110
           db   %01100010
           db   %11000000
           db   %11001110
           db   %11000110
           db   %01100110
           db   %00110110
           db   %00000000
; "H" (143)
           db   %11110110
           db   %01100110
           db   %01100110
           db   %01101110
           db   %01100110
           db   %01100110
           db   %11110110
           db   %00000000
; "I" (151)
           db   %00111100
           db   %00011000
           db   %00011000
           db   %00011000
           db   %00011000
           db   %00011000
           db   %00111100
           db   %00000000
; "L" (159)
           db   %11110000
           db   %01100000
           db   %01100000
           db   %01100000
           db   %01100000
           db   %01100110
           db   %11101110
           db   %00000000
; "M" (167)
           db   %01000110
           db   %01101110
           db   %00111110
           db   %01010110
           db   %01000110
           db   %01000110
           db   %11101110
           db   %00000000
; "N" (175)
           db   %01000110
           db   %01100010
           db   %00110010
           db   %01011010
           db   %01001110
           db   %01000110
           db   %11100010
           db   %00000000
; "O" (183)
           db   %01101100
           db   %11000110
           db   %11000110
           db   %11000110
           db   %11000110
           db   %11000110
           db   %01101100
           db   %00000000
; "P" (191)
           db   %11101100
           db   %01100110
           db   %01100110
           db   %01101100
           db   %01100000
           db   %01100000
           db   %11110000
           db   %00000000
; "R" (199)
           db   %11101100
           db   %01100110
           db   %01100110
           db   %01101100
           db   %01101100
           db   %01100110
           db   %11100110
           db   %00000000
; "S" (207)
           db   %01101110
           db   %11000110
           db   %11100000
           db   %01111100
           db   %00001110
           db   %11000110
           db   %11101100
           db   %00000000
; "T" (215)
           db   %01111110
           db   %01011010
           db   %00011000
           db   %00011000
           db   %00011000
           db   %00011000
           db   %00111100
           db   %00000000
; "U" (223)
           db   %11110110
           db   %01100010
           db   %01100010
           db   %01100010
           db   %01100010
           db   %01100010
           db   %00110100
           db   %00000000
; "V" (231)
           db   %11110110
           db   %01100010
           db   %01100010
           db   %00110100
           db   %00110100
           db   %00011000
           db   %00011000
           db   %00000000
; "Y" (239)
           db   %11110110
           db   %01100010
           db   %01110100
           db   %00111000
           db   %00011000
           db   %00011000
           db   %00111100
           db   %00000000
; "(C)" (247)
           db   %00111100
           db   %01000010
           db   %10011001
           db   %10100001
           db   %10100001
           db   %10011001
           db   %01000010
           db   %00111100
; " " (255)
; Blank character - initialized by code

; ######################### Bit patterns for sprites ###########################

; 00 = Screen color (Blue)
; 01 = Sprite multicolor #0 (Grey)
; 10 = Sprite color (Yellow)
; 11 = Sprite multicolor #1 (Light grey)

           db   new_addr,lo tmpspdata,hi tmpspdata
; Left part of ship moving right
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000011
           db   %00000000
           db   %01010101
           db   %00111111
           db   %00111111
           db   %00001111
           db   %00001111
           db   %00000001

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00010000
           db   %11010000
           db   %01010000
           db   %01010101
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %01010101

           db   %00000111
           db   %00000001
           db   %00000001
           db   %00011111
           db   %00011111
           db   %00011111
           db   %00011111
           db   %00011111
           db   %00011111
           db   %01011111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %01010101

; Right part of ship moving right
           db   %11000000
           db   %00000000
           db   %00000000
           db   %11000000
           db   %11000000
           db   %11000000
           db   %11110000
           db   %11110000
           db   %11110000
           db   %11010111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %01010101

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00010100
           db   %00010111
           db   %00010101
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %01010101

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %11110000
           db   %00000000
           db   %11111111
           db   %11111111
           db   %11111100
           db   %11111100
           db   %11110000
           db   %11110000
           db   %01000000

; Left part of ship on fire frame #1
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000011
           db   %00000000
           db   %01010101
           db   %00111110
           db   %00111111
           db   %00001111
           db   %00001111
           db   %00000001

           db   %10000000
           db   %00100000
           db   %00100000
           db   %10101000
           db   %00101000
           db   %00101010
           db   %00011010
           db   %11011010
           db   %01101010
           db   %10101010
           db   %10101011
           db   %10101011
           db   %11101111
           db   %11111111
           db   %01010101

           db   %00000111
           db   %00000010
           db   %00000001
           db   %00011111
           db   %00011110
           db   %00011110
           db   %00011111
           db   %10011111
           db   %10011110
           db   %01011010
           db   %11111010
           db   %11111010
           db   %11111110
           db   %11111111
           db   %01010101

; Right part of ship on fire frame #1
           db   %11000000
           db   %00000000
           db   %10000000
           db   %10000010
           db   %10000000
           db   %10000000
           db   %10100000
           db   %10100000
           db   %10110000
           db   %10100111
           db   %10101111
           db   %10111111
           db   %11111111
           db   %11111111
           db   %01010101

           db   %00001000
           db   %00100000
           db   %10100000
           db   %10100000
           db   %10101000
           db   %00101000
           db   %00101010
           db   %00011010
           db   %11111010
           db   %11101011
           db   %10101010
           db   %11101010
           db   %11111011
           db   %11111111
           db   %01010101

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %11110000
           db   %10000000
           db   %11111111
           db   %11111111
           db   %11111100
           db   %11111100
           db   %11110000
           db   %11110000
           db   %01000000

; Left part of ship on fire frame #2
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000011
           db   %00000000
           db   %01010110
           db   %00111111
           db   %00111111
           db   %00001111
           db   %00001111
           db   %00000001

           db   %00000000
           db   %00100000
           db   %00001000
           db   %00101000
           db   %10100000
           db   %10100000
           db   %00101000
           db   %10101000
           db   %10101010
           db   %10101010
           db   %10101011
           db   %10101011
           db   %11101111
           db   %11111111
           db   %01010101

           db   %00000111
           db   %00000001
           db   %00000010
           db   %00011110
           db   %00011010
           db   %00011010
           db   %00011110
           db   %00011010
           db   %00011010
           db   %01011110
           db   %11111110
           db   %11111010
           db   %11111110
           db   %11111111
           db   %01010101

; Right part of ship on fire frame #2
           db   %11000000
           db   %00000000
           db   %00000000
           db   %11000000
           db   %11000000
           db   %11000000
           db   %10110000
           db   %10110000
           db   %10100000
           db   %10100111
           db   %10111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %01010101

           db   %00000000
           db   %00001000
           db   %00001000
           db   %00101000
           db   %00001010
           db   %00011010
           db   %10101011
           db   %10101001
           db   %11101010
           db   %11101010
           db   %11101011
           db   %11111010
           db   %11111011
           db   %11111111
           db   %01010101

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %11110000
           db   %00000000
           db   %11111111
           db   %11111111
           db   %11111100
           db   %11111100
           db   %11110000
           db   %11110000
           db   %01000000

; Left part of sub moving right
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00110101
           db   %11111111
           db   %11111111
           db   %01111111
           db   %01111111
           db   %00011111
           db   %00000101

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %01010101
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %01010101

           db   %00110000
           db   %00110000
           db   %00110000
           db   %00110000
           db   %01111111
           db   %01111111
           db   %01111111
           db   %01111111
           db   %01111101
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %01010101

; Right part of sub moving right
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000111
           db   %00000001
           db   %01010101
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %01010101

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %11110000
           db   %00000000
           db   %01010101
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %01010101

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %01011100
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %11111100
           db   %01010000

; Left part of sub explosion frame #2
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00110101
           db   %11111011
           db   %11101010
           db   %10101010
           db   %01101010
           db   %00011011
           db   %00000101

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000010
           db   %01010110
           db   %11111110
           db   %11111110
           db   %10111111
           db   %11111111
           db   %11111111
           db   %01010101

           db   %00110000
           db   %00110000
           db   %00110000
           db   %00110000
           db   %01111111
           db   %01101010
           db   %10101010
           db   %10101010
           db   %10101010
           db   %10101010
           db   %10101010
           db   %10101010
           db   %11101010
           db   %11111111
           db   %01010101

; Right part of sub explosion frame #2
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %10101000
           db   %10101010
           db   %10101010
           db   %10101010
           db   %10101010
           db   %10101010
           db   %10101010
           db   %10101011
           db   %11111111
           db   %01010101

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %11110000
           db   %10000000
           db   %10010101
           db   %10111111
           db   %10111111
           db   %11111111
           db   %11111111
           db   %11111111
           db   %01010101

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %01011100
           db   %11101111
           db   %10101011
           db   %11101111
           db   %11111111
           db   %11111100
           db   %01010000

; Left part of sub explosion frame #3
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00001000
           db   %00101010
           db   %10101010
           db   %10100010
           db   %10100010
           db   %00101010
           db   %00101010
           db   %00001000

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000010
           db   %00001010
           db   %00001010
           db   %00101010
           db   %00101010
           db   %00001010
           db   %10001010
           db   %10001110
           db   %10111111
           db   %11110011
           db   %01010000

           db   %00000000
           db   %00000010
           db   %00111010
           db   %10101010
           db   %10101010
           db   %10101010
           db   %10101100
           db   %10100000
           db   %10100000
           db   %10101000
           db   %10101010
           db   %10101010
           db   %11101010
           db   %11111010
           db   %01010010

; Right part of sub explosion frame #3
           db   %00000000
           db   %10000000
           db   %10100000
           db   %10101010
           db   %10101010
           db   %10101010
           db   %00101010
           db   %00001010
           db   %00001010
           db   %00111010
           db   %10101010
           db   %10101010
           db   %10101011
           db   %10101111
           db   %11010100

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %10000000
           db   %10100000
           db   %10100000
           db   %10101000
           db   %10101000
           db   %10101111
           db   %10111110
           db   %11111110
           db   %11001111
           db   %00111111
           db   %00010101

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00100000
           db   %10101000
           db   %10101000
           db   %10001010
           db   %10001010
           db   %10101010
           db   %10101000
           db   %01100000

; Left part of sub explosion frame #4
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00001000
           db   %00100010
           db   %10000011
           db   %10000000
           db   %10110000
           db   %00100010
           db   %00001010
           db   %00000000

           db   %00000000
           db   %00000000
           db   %00000010
           db   %00001010
           db   %00001010
           db   %00101011
           db   %00101100
           db   %00100000
           db   %00101000
           db   %00101000
           db   %00001011
           db   %10111010
           db   %10000010
           db   %00000010
           db   %00000000

           db   %00000000
           db   %00101010
           db   %10101010
           db   %10100000
           db   %11000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %10000000
           db   %10110000
           db   %10101010
           db   %00101010

; Right part of sub explosion frame #4
           db   %00000000
           db   %10101000
           db   %10101010
           db   %00001010
           db   %00000011
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000010
           db   %00001110
           db   %00111010
           db   %10101000
           db   %10000000

           db   %00000000
           db   %00000000
           db   %10000000
           db   %10100000
           db   %10100000
           db   %10101000
           db   %00101000
           db   %00101000
           db   %00101000
           db   %10101011
           db   %10100000
           db   %10110000
           db   %00001100
           db   %00000000
           db   %00000000

           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00000000
           db   %00100000
           db   %00101000
           db   %10001110
           db   %10000010
           db   %00000010
           db   %00001000
           db   %10001000
           db   %10100000
           db   %00100000

           db   endofdata
